# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2017-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import os
import os.path
import shutil
import sys

# This function just adds an extra zero when the hex number as char has only
# single digit
def hex0x0to0x00(inp):
    if inp == "0x0":
        return "0x00"
    if inp == "0x1":
        return "0x01"
    if inp == "0x2":
        return "0x02"
    if inp == "0x3":
        return "0x03"
    if inp == "0x4":
        return "0x04"
    if inp == "0x5":
        return "0x05"
    if inp == "0x6":
        return "0x06"
    if inp == "0x7":
        return "0x07"
    if inp == "0x8":
        return "0x08"
    if inp == "0x9":
        return "0x09"
    if inp == "0xa":
        return "0x0a"
    if inp == "0xb":
        return "0x0b"
    if inp == "0xc":
        return "0x0c"
    if inp == "0xd":
        return "0x0d"
    if inp == "0xe":
        return "0x0e"
    if inp == "0xf":
        return "0x0f"
    return inp



def main():
    sipBin = ""
    sipHeader = ""
    sipCharArrayName = ""
    if len(sys.argv) > 3 :
        sipBin = sys.argv[1]
        sipHeader = sys.argv[2]
        sipCharArrayName = sys.argv[3]
        if os.path.exists(sipBin):
            sipBinArray = open(sipBin,"rb").read()
            sipHeaderFile = open(sipHeader,"w+")
            print("Found "+sipBin)
            print("Converting "+str(sipBin)+" to "+str(sipHeader))
            sipHeaderFile.write("namespace SIP\n")
            sipHeaderFile.write("{\n")
            sipHeaderFile.write("    const unsigned char "+sipCharArrayName+"[] =\n")
            sipHeaderFile.write("    {\n")
            lenA = len(sipBinArray)
            for i in range(lenA):
                if i >0 and i % 16 == 0 :
                    sipHeaderFile.write("\n")
                sipHeaderFile.write(hex0x0to0x00(hex(sipBinArray[i])).lower())
                if i+1 < lenA:
                    sipHeaderFile.write(", ")
            sipHeaderFile.write("};\n}")
            print("Conversion  done.. \n")
            sipHeaderFile.close()
        else:
            print("Error : Unable to open "+sipInput)
    else:
        print("Usage sip.py sip.bin sip.h sipStruct")
    return

if __name__ == "__main__":
    main()


# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2020-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

# -*- Python -*-

import lit.formats
import lit.util
import os

from lit.llvm import llvm_config
from lit.llvm.subst import ToolSubst
from lit.llvm.subst import FindTool

# Configuration file for the 'lit' test runner.

# name: The name of this test suite.
config.name = 'vc-opt'

# testFormat: The test format to use to interpret tests.
config.test_format = lit.formats.ShTest(not llvm_config.use_lit_shell)

# suffixes: A list of file extensions to treat as test files.
config.suffixes = ['.ll']

# excludes: A list of directories  and files to exclude from the testsuite.
config.excludes = ['CMakeLists.txt']

# test_source_root: The root path where tests are located.
config.test_source_root = os.path.dirname(__file__)

# test_exec_root: The root path where tests should be run.
config.test_exec_root = os.path.join(config.test_run_dir, 'test_output')

llvm_config.with_environment('LD_LIBRARY_PATH', config.cm_opt_lib_dir, append_path=True)

llvm_config.use_default_substitutions()

config.substitutions.append(('%PATH%', config.environment['PATH']))
config.substitutions.append(('%OCL_GENERIC_BIF%', config.ocl_generic_bif))
config.substitutions.append(('%VC_PRITF_OCL_BIF%', config.vc_printf_ocl_bif))
config.substitutions.append(('%VC_SPIRV_OCL_BIF%', config.vc_spirv_ocl_bif))

platforms = config.vc_platform_list.split(";")
for platform in platforms:
  builtins_path = '{}_{}.vccg.bc'.format(config.vc_builtins_bif_prefix, platform)
  config.substitutions.append(('%VC_BUILTINS_BIF_{}%'.format(platform), builtins_path))

if config.use_khronos_spirv_translator_in_sc == "1":
  config.substitutions.append(('%SPV_CHECK_PREFIX%', 'CHECK-KHR'))
else:
  config.substitutions.append(('%SPV_CHECK_PREFIX%', 'CHECK-LEGACY'))

tool_dirs = [
  config.cm_opt_bin_dir,
  config.oneapi_readelf_dir,
  config.llvm_tools_dir]

# Add extra args for opt to remove boilerplate from tests.
vc_extra_args = ['-load', config.llvm_plugin]
tools = [ToolSubst('not'),
         ToolSubst('opt', extra_args=vc_extra_args+[config.llvm_dependent_opt_flags]),
         ToolSubst('llc', extra_args=vc_extra_args),
         ToolSubst('oneapi-readelf', unresolved='ignore'),
         ToolSubst('llvm-dwarfdump'),
         ToolSubst('%igc-lld', command=FindTool('ld.lld'))]

if int(config.llvm_version) < 11:
  config.substitutions.append(('%not_for_vc_diag%', 'not'))
  config.substitutions.append(('%use_old_pass_manager%', ''))
else:
  config.substitutions.append(('%not_for_vc_diag%', 'not --crash'))
  config.substitutions.append(('%use_old_pass_manager%', '-enable-new-pm=0'))

if int(config.llvm_version) < 12:
  config.available_features.add('llvm_11_or_less')
else:
  config.available_features.add('llvm_12_or_greater')

if config.oneapi_readelf_dir:
  config.available_features.add('oneapi-readelf')

llvm_config.add_tool_substitutions(tools, tool_dirs)



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import os
import argparse
import re

from pathlib import Path
from mako.template import Template
from mako.lookup import TemplateLookup

def safe_open(path, mode = 'w'):
    os.makedirs(os.path.dirname(path), exist_ok=True)
    return open(path, mode)

def remove_all_whitespace_at_line_end(text):
    """Removes all whitespace at the ends of lines
    Args:
        text:   An input text.
    Returns:
        The formatted text without trailing whitespaces at the ends of lines"""
    text = re.sub("[ \t]+(\r?)\n", '\g<1>\n', text, flags=re.MULTILINE)
    return text

def set_windows_line_ends(text):
    """Removes all whitespace at the ends of lines
    Args:
        text:   An input text.
    Returns:
        The formatted text without trailing whitespaces at the ends of lines"""
    text = re.sub("(?<!\r)\n", "\r\n", text, flags=re.MULTILINE)
    return text

def write_to_file_using_template(file_path, template, **kwargs):
    """Gets the formatted text with intendation which corresponds to 4 spaces.
    Args:
        file_path:   output path,
        template:    recipe for a new file,
        kwargs:      arguments for template."""
    try:
        print(" --- Generating %s" % os.path.basename(file_path))
        content = template.render(file_name=os.path.basename(file_path), **kwargs)
        content = remove_all_whitespace_at_line_end(content)
        content = set_windows_line_ends(content)
        if os.path.isfile(file_path):
            os.chmod(file_path, 0o666)
        with safe_open(file_path, 'wb') as f:
            f.write(content.encode())
        print(" --- Successfully generated in the following location: \n    %s"%(os.path.abspath(file_path)))
        os.chmod(file_path, 0o444)
    except IOError as e:
        print(" --- The file is not generated in the following location: \n    %s"%(os.path.abspath(file_path)))
        print("I/O error: %s" % e.strerror)
    except:
        print(" --- The file is not generated in the following location: \n    %s"%(os.path.abspath(file_path)))
        print(exceptions.text_error_template().render())

def from_template_name_to_destination_name(template_filename : str):
    return Path(template_filename).name.replace(".mako", "")

def file_path(path):
    if os.path.isfile(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"{path} is not a valid path to file")

def dir_path(path):
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"{path} is not a valid path to dir")



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2022-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

#===----------------------------------------------------------------------===//
#
# This file defines all of the Internal-specific intrinsics, which correspond to
# vISA instructions.
#
# Comment lines with a triple slash ### introduction are extracted and
# appended to docs/Targets/Internal/InternalLangRef.rst to give the Internal backend
# language reference in docs/autogenerated/Targets/Internal/InternalLangRef.rst.
#
#===------------------------------------------------------------------------===#

#------------ Currently Supported Types ----------------------
#PointerTypes = ["ptr_private", "ptr_global", "ptr_constant", "ptr_local", "ptr_generic"]
#FloatingPointTypes = ["half", "float", "double"]
#IntegerTypes = ["bool", "char", "short", "int", "long"]
#AdditionalTypes = ["vararg"]
#IntrinsicsProperties = ["None", "NoMem", "ReadArgMem", "ReadMem", "ReadWriteArgMem", "NoReturn", "NoDuplicate", "Convergent"]
#IntrinsicsProperties may be specified as a comma separated list(e.g., "Convergent,NoMem")
#
# EX. "blah": {"result" : {return_type}, "arguments" : [arg1_type, arg2_type.....], "attributes" : Property }
#
# The "any" type can be followed by a default type if a type is not explicitly specified : Ex. "any:int"
#
# 0 - LLVMMatchType<0>
# 1 - LLVMMatchType<1>
# {int} - LLVMMatchType<{int}>

Imported_Intrinsics = {
## ``llvm.vc.internal.jump.table`` : CMC internal, no VISA
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: integer BasicBlock index in the full set of destinations
## * arg1-N: the full set of switch labels
##
## * Return value: selected label
##
## The intrinsic is a helper for switch jump tables generation. Arg0
## will be used by visa switchjmp as index. Return value and arg1-N are
## used to make ir semantically legal.
##
    "jump_table" : { "result" : "anyptr",
                     "arguments" :  ["anyint", "vararg"],
                     "attributes" :  "NoMem" },

## ``llvm.vc.internal.read.variable.region`` : read a vISA variable region
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: ptr pointer to a global variable that corresponds to a vISA variable
##         (overloaded)
## * arg1: i32 vstride in elements, constant
## * arg2: i32 width in elements, constant
## * arg3: i32 stride in elements, constant
## * arg4: i32 offset in elements, constant
##
## * Return value: iN, fN, vXiN, vXfN the read value (overloaded)
##
## This corresponds to MOV instruction or a general source operand in visa.
## Utilizes technique of using global variable in LLVM IR for predefined
## vISA variables.
##
    "read_variable_region" : { "result": "any",
                               "arguments" : ["anyptr", "int", "int", "int",
                                              "int"],
                               "attributes" : "ReadMem", },

## ``llvm.vc.internal.write.variable.region`` : write a vISA variable region
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: ptr pointer to a global variable that corresponds to a vISA variable
##         (overloaded)
## * arg1: iN, fN, vXiN, vXfN value to write (overloaded)
## * arg2: i32 stride in elements, constant
## * arg3: i32 offset in elements, constant
## * arg4: i1 or vXi1 mask (overloaded)
##
## This corresponds to MOV instruction or a general destination operand in visa.
## Utilizes technique of using global variable in LLVM IR for predefined
## vISA variables.
##
    "write_variable_region" : { "result": "void",
                                "arguments" : ["anyptr", "any", "int",
                                               "int", "anyint"],
                                "attributes" : "WriteMem", },

## ``llvm.vc.internal.cast.to.ptr.explicit`` : convert ptr_generic to
## private/local/global ptr.
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: generic pointer
##
## * Return value: private/local/global pointer
##
## This intrisic attempts to explicitly convert a generic ptr to a
##  private/local/global ptr. If the cast fails the intrisic returns null pointer.
    "cast_to_ptr_explicit" : { "result": "anyptr",
                               "arguments": ["ptr_generic"],
                               "attributes": "NoMem", },

### --------------
### ALU intrinsics
### --------------

## ``llvm.vc.internal.cast.to.bf16`` : convert float into bfloat16
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: input data, f32 scalar or vector (overloaded)
##
## * Return value: i16 scalar or vector (overloaded)
##
## This intrinsic represents float -> bfloat16 conversion operation
    "cast_to_bf16" : { "result": "anyint",
                       "arguments": ["anyfloat"],
                       "attributes": "NoMem", },
## ``llvm.vc.internal.cast.from.bf16`` : convert bfloat16 into float
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: bfloat16 input data, i16 scalar or vector (overloaded)
##
## * Return value: f32 scalar or vector (overloaded)
##
## This intrinsic represents float -> bfloat16 conversion operation
    "cast_from_bf16" : { "result": "anyfloat",
                         "arguments": ["anyint"],
                         "attributes": "NoMem", },

## ``llvm.vc.internal.round.to.tf32`` : round float into tfloat32
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: input data, f32 scalar or vector (overloaded)
##
## * Return value: i32 scalar or vector (overloaded)
##
## This intrinsic represents float -> tfloat32 conversion operation
    "round_to_tf32" : { "result": "anyfloat",
                        "arguments": ["anyint"],
                        "attributes": "NoMem", },

## ``llvm.vc.internal.stochastic.round.to.f16`` : half stochastic rounding operation
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: input data, f32 scalar or vector (overloaded)
## * arg1: random number, i16 scalar or vector of the same width as arg0
##
## * Return value: f16 scalar or vector of the same width as arg0
##
## This intrinsic represents float -> half stochastic rounding operation
    "stochastic_round_to_f16" : { "result": "anyfloat",
                                  "arguments": ["anyfloat", "anyint"],
                                  "attributes": "NoMem", },

### ---------------------------
### Low-level memory intrinsics
### ---------------------------

## ``llvm.vc.internal.lsc.atomic.*``: LSC atomic intrinsics
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: vNxi1 Predicate (overloaded)
## * arg1: i8 Atomic opcode [MBC]
## * arg2: i8 Address size [MBC]
## * arg3: i8 Element size [MBC]
## * arg4: i8 L1 cache controls [MBC]
## * arg5: i8 L3 cache controls [MBC]
## * arg6: i64 Address base (for stateless)
##         i32 BTI (for stateful)
## * arg7: vNxi32 or vNxi64 Address indices (overloaded)
## * arg8: i16 Address scale [MBC]
## * arg9: i32 Address immediate offset [MBC]
## * arg10: 1st source vector for the atomic operation,
##          must be undef for unary operations
## * arg11: 2nd source vector for the atomic operation,
##          must be undef for unary and binary operations
## * arg12: vector to take values for masked simd lanes from
##
## * Return value: the value read from memory, merged with arg12 by predicate
##
    "lsc_atomic_bti": { "result": "anyvector",
                        "arguments": [
                            "anyint", # vNxi1, predicate
                            "char",   # atomic opcode
                            "char",   # address size
                            "char",   # element size
                            "char",   # L1 cache control
                            "char",   # L3 cache control
                            "int",    # i32 BTI
                            "anyint", # vNi32 address offsets
                            "short",  # address scale
                            "int",    # address immediate offset
                            0,        # src1
                            0,        # src2
                            0,        # passthru value
                        ],
                        "attributes": "SideEffects", },
    "lsc_atomic_bss": { "result": "anyvector",
                        "arguments": [
                            "anyint", # vNxi1, predicate
                            "char",   # atomic opcode
                            "char",   # address size
                            "char",   # element size
                            "char",   # L1 cache control
                            "char",   # L3 cache control
                            "int",    # i32 BSS
                            "anyint", # vNi32 address offsets
                            "short",  # address scale
                            "int",    # address immediate offset
                            0,        # src1
                            0,        # src2
                            0,        # passthru value
                        ],
                        "attributes": "SideEffects", },
    "lsc_atomic_slm": { "result": "anyvector",
                        "arguments": [
                            "anyint", # vNxi1, predicate
                            "char",   # atomic opcode
                            "char",   # address size
                            "char",   # element size
                            "char",   # L1 cache control
                            "char",   # L3 cache control
                            "int",    # i32 address base
                            "anyint", # vNi32 address offsets
                            "short",  # address scale
                            "int",    # address immediate offset
                            0,        # src1
                            0,        # src2
                            0,        # passthru value
                        ],
                      "attributes": "SideEffects", },
    "lsc_atomic_ugm": { "result": "anyvector",
                        "arguments": [
                            "anyint", # vNxi1, predicate
                            "char",   # atomic opcode
                            "char",   # address size
                            "char",   # element size
                            "char",   # L1 cache control
                            "char",   # L3 cache control
                            "long",   # i64 address base
                            "anyint", # vNi32 or vNi64 address offsets
                            "short",  # address scale
                            "int",    # address immediate offset
                            0,        # src1
                            0,        # src2
                            0,        # passthru value
                        ],
                        "attributes": "SideEffects", },

## ``llvm.vc.internal.lsc.load.*`` : LSC load intrinsics
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: vNxi1 Predicate (overloaded)
## * arg1: i8 Address size [MBC]
## * arg2: i8 Element size [MBC]
## * arg3: i8 Vector size (number of elements per SIMD lane) [MBC]
##         i8 Channel mask (for quad intrinsics) [MBC]
## * arg4: i8 L1 cache controls [MBC]
## * arg5: i8 L3 cache controls [MBC]
## * arg6: i64 Address base (for stateless)
##         i32 BTI (for stateful)
## * arg7: vNxi32 or vNxi64 Address indices (overloaded)
## * arg8: i16 Address scale [MBC]
## * arg9: i32 Address immediate offset [MBC]
## * arg10: vector to take values for masked simd lanes from
##
## * Return value: the value read from memory, merged with arg10 by predicate
##
    "lsc_load_bti": { "result": "anyvector",
                      "arguments": [
                          "anyint", # vNxi1, predicate
                          "char",   # address size
                          "char",   # element size
                          "char",   # vector size
                          "char",   # L1 cache control
                          "char",   # L3 cache control
                          "int",    # i32 BTI
                          "anyint", # vNi32 address offsets
                          "short",  # address scale
                          "int",    # address immediate offset
                          0,        # passthru value
                      ],
                      "attributes": "ReadMem", },
    "lsc_load_bss": { "result": "anyvector",
                      "arguments": [
                          "anyint", # vNxi1, predicate
                          "char",   # address size
                          "char",   # element size
                          "char",   # vector size
                          "char",   # L1 cache control
                          "char",   # L3 cache control
                          "int",    # i32 BSS
                          "anyint", # vNi32 address offsets
                          "short",  # address scale
                          "int",    # address immediate offset
                          0,        # passthru value
                      ],
                      "attributes": "ReadMem", },
    "lsc_load_slm": { "result": "anyvector",
                      "arguments": [
                          "anyint", # vNxi1, predicate
                          "char",   # address size
                          "char",   # element size
                          "char",   # vector size
                          "char",   # L1 cache control
                          "char",   # L3 cache control
                          "int",    # i32 address base
                          "anyint", # vNi32 address offsets
                          "short",  # address scale
                          "int",    # address immediate offset
                          0,        # passthru value
                      ],
                      "attributes": "ReadMem", },
    "lsc_load_ugm": { "result": "anyvector",
                      "arguments": [
                          "anyint", # vNxi1, predicate
                          "char",   # address size
                          "char",   # element size
                          "char",   # vector size
                          "char",   # L1 cache control
                          "char",   # L3 cache control
                          "long",   # i64 address base
                          "anyint", # vNi32 or vNi64 address offsets
                          "short",  # address scale
                          "int",    # address immediate offset
                          0,        # passthru value
                      ],
                      "attributes": "ReadMem", },

    "lsc_load_quad_bti": { "result": "anyvector",
                           "arguments": [
                               "anyint", # vNxi1, predicate
                               "char",   # address size
                               "char",   # element size
                               "char",   # channel mask
                               "char",   # L1 cache control
                               "char",   # L3 cache control
                               "int",    # i32 BTI
                               "anyint", # vNi32 address offsets
                               "short",  # address scale
                               "int",    # address immediate offset
                               0,        # passthru value
                           ],
                           "attributes": "ReadMem", },
    "lsc_load_quad_bss": { "result": "anyvector",
                           "arguments": [
                               "anyint", # vNxi1, predicate
                               "char",   # address size
                               "char",   # element size
                               "char",   # channel mask
                               "char",   # L1 cache control
                               "char",   # L3 cache control
                               "int",    # i32 BSS
                               "anyint", # vNi32 address offsets
                               "short",  # address scale
                               "int",    # address immediate offset
                               0,        # passthru value
                           ],
                           "attributes": "ReadMem", },
    "lsc_load_quad_slm": { "result": "anyvector",
                           "arguments": [
                               "anyint", # vNxi1, predicate
                               "char",   # address size
                               "char",   # element size
                               "char",   # channel mask
                               "char",   # L1 cache control
                               "char",   # L3 cache control
                               "int",    # i32 address base
                               "anyint", # vNi32 address offsets
                               "short",  # address scale
                               "int",    # address immediate offset
                               0,        # passthru value
                           ],
                           "attributes": "ReadMem", },
    "lsc_load_quad_ugm": { "result": "anyvector",
                           "arguments": [
                               "anyint", # vNxi1, predicate
                               "char",   # address size
                               "char",   # element size
                               "char",   # channel mask
                               "char",   # L1 cache control
                               "char",   # L3 cache control
                               "long",   # i64 address base
                               "anyint", # vNi32 or vNi64 address offsets
                               "short",  # address scale
                               "int",    # address immediate offset
                               0,        # passthru value
                           ],
                           "attributes": "ReadMem", },

## ``llvm.vc.internal.lsc.prefetch.*`` : LSC prefetch intrinsics
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: vNxi1 Predicate (overloaded)
## * arg1: i8 Address size [MBC]
## * arg2: i8 Element size [MBC]
## * arg3: i8 Vector size (number of elements per SIMD lane) [MBC]
##         i8 Channel mask (for quad intrinsics) [MBC]
## * arg4: i8 L1 cache controls [MBC]
## * arg5: i8 L3 cache controls [MBC]
## * arg6: i64 Address base (for stateless)
##         i32 BTI (for stateful)
## * arg7: vNxi32 or vNxi64 Address indices (overloaded)
## * arg8: i16 Address scale [MBC]
## * arg9: i32 Address immediate offset [MBC]
##
## * Return value: void
##
    "lsc_prefetch_bti": { "result": "void",
                          "arguments": [
                              "anyint", # vNxi1, predicate
                              "char",   # address size
                              "char",   # element size
                              "char",   # vector size
                              "char",   # L1 cache control
                              "char",   # L3 cache control
                              "int",    # i32 BTI
                              "anyint", # vNi32 address offsets
                              "short",  # address scale
                              "int",    # address immediate offset
                          ],
                          "attributes": "SideEffects", },
    "lsc_prefetch_bss": { "result": "void",
                          "arguments": [
                              "anyint", # vNxi1, predicate
                              "char",   # address size
                              "char",   # element size
                              "char",   # vector size
                              "char",   # L1 cache control
                              "char",   # L3 cache control
                              "int",    # i32 BSS
                              "anyint", # vNi32 address offsets
                              "short",  # address scale
                              "int",    # address immediate offset
                          ],
                          "attributes": "SideEffects", },
    "lsc_prefetch_ugm": { "result": "void",
                          "arguments": [
                              "anyint", # vNxi1, predicate
                              "char",   # address size
                              "char",   # element size
                              "char",   # vector size
                              "char",   # L1 cache control
                              "char",   # L3 cache control
                              "long",   # i64 address base
                              "anyint", # vNi32 or vNi64 address offsets
                              "short",  # address scale
                              "int",    # address immediate offset
                          ],
                          "attributes": "SideEffects", },

    "lsc_prefetch_quad_bti": { "result": "void",
                               "arguments": [
                                   "anyint", # vNxi1, predicate
                                   "char",   # address size
                                   "char",   # element size
                                   "char",   # channel mask
                                   "char",   # L1 cache control
                                   "char",   # L3 cache control
                                   "int",    # i32 BTI
                                   "anyint", # vNi32 address offsets
                                   "short",  # address scale
                                   "int",    # address immediate offset
                               ],
                               "attributes": "SideEffects", },
    "lsc_prefetch_quad_bss": { "result": "void",
                               "arguments": [
                                   "anyint", # vNxi1, predicate
                                   "char",   # address size
                                   "char",   # element size
                                   "char",   # channel mask
                                   "char",   # L1 cache control
                                   "char",   # L3 cache control
                                   "int",    # i32 BSS
                                   "anyint", # vNi32 address offsets
                                   "short",  # address scale
                                   "int",    # address immediate offset
                               ],
                               "attributes": "SideEffects", },
    "lsc_prefetch_quad_ugm": { "result": "void",
                               "arguments": [
                                   "anyint", # vNxi1, predicate
                                   "char",   # address size
                                   "char",   # element size
                                   "char",   # channel mask
                                   "char",   # L1 cache control
                                   "char",   # L3 cache control
                                   "long",   # i64 address base
                                   "anyint", # vNi32 or vNi64 address offsets
                                   "short",  # address scale
                                   "int",    # address immediate offset
                               ],
                               "attributes": "SideEffects", },

## ``llvm.vc.internal.lsc.store.*`` : LSC store intrinsics
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * arg0: vNxi1 Predicate (overloaded)
## * arg1: i8 Address size [MBC]
## * arg2: i8 Element size [MBC]
## * arg3: i8 Vector size (number of elements per SIMD lane) [MBC]
##         i8 Channel mask (for quad intrinsics) [MBC]
## * arg4: i8 L1 cache controls [MBC]
## * arg5: i8 L3 cache controls [MBC]
## * arg6: i64 Address base (for stateless)
##         i32 BTI (for stateful)
## * arg7: vNxi32 or vNxi64 Address indices (overloaded)
## * arg8: i16 Address scale [MBC]
## * arg9: i32 Address immediate offset [MBC]
## * arg10: Data to write (overloaded)
##
## * Return value: void
##
    "lsc_store_bti": { "result": "void",
                       "arguments": [
                           "anyint", # vNxi1, predicate
                           "char",   # address size
                           "char",   # element size
                           "char",   # vector size
                           "char",   # L1 cache control
                           "char",   # L3 cache control
                           "int",    # i32 BTI
                           "anyint", # vNi32 address offsets
                           "short",  # address scale
                           "int",    # address immediate offset
                           "anyvector", # Data to write
                       ],
                       "attributes": "WriteMem", },
    "lsc_store_bss": { "result": "void",
                       "arguments": [
                           "anyint", # vNxi1, predicate
                           "char",   # address size
                           "char",   # element size
                           "char",   # vector size
                           "char",   # L1 cache control
                           "char",   # L3 cache control
                           "int",    # i32 BSS
                           "anyint", # vNi32 address offsets
                           "short",  # address scale
                           "int",    # address immediate offset
                           "anyvector", # Data to write
                       ],
                       "attributes": "WriteMem", },
    "lsc_store_slm": { "result": "void",
                       "arguments": [
                           "anyint", # vNxi1, predicate
                           "char",   # address size
                           "char",   # element size
                           "char",   # vector size
                           "char",   # L1 cache control
                           "char",   # L3 cache control
                           "int",    # i32 address base
                           "anyint", # vNi32 address offsets
                           "short",  # address scale
                           "int",    # address immediate offset
                           "anyvector", # Data to write
                       ],
                       "attributes": "WriteMem", },
    "lsc_store_ugm": { "result": "void",
                       "arguments": [
                           "anyint", # vNxi1, predicate
                           "char",   # address size
                           "char",   # element size
                           "char",   # vector size
                           "char",   # L1 cache control
                           "char",   # L3 cache control
                           "long",   # i64 address base
                           "anyint", # vNi32 or vNi64 address offsets
                           "short",  # address scale
                           "int",    # address immediate offset
                           "anyvector", # Data to write
                       ],
                       "attributes": "WriteMem", },

    "lsc_store_quad_bti": { "result": "void",
                            "arguments": [
                                "anyint", # vNxi1, predicate
                                "char",   # address size
                                "char",   # element size
                                "char",   # channel mask
                                "char",   # L1 cache control
                                "char",   # L3 cache control
                                "int",    # i32 BTI
                                "anyint", # vNi32 address offsets
                                "short",  # address scale
                                "int",    # address immediate offset
                                "anyvector", # Data to write
                            ],
                            "attributes": "WriteMem", },
    "lsc_store_quad_bss": { "result": "void",
                            "arguments": [
                                "anyint", # vNxi1, predicate
                                "char",   # address size
                                "char",   # element size
                                "char",   # channel mask
                                "char",   # L1 cache control
                                "char",   # L3 cache control
                                "int",    # i32 BSS
                                "anyint", # vNi32 address offsets
                                "short",  # address scale
                                "int",    # address immediate offset
                                "anyvector", # Data to write
                            ],
                            "attributes": "WriteMem", },
    "lsc_store_quad_slm": { "result": "void",
                            "arguments": [
                                "anyint", # vNxi1, predicate
                                "char",   # address size
                                "char",   # element size
                                "char",   # channel mask
                                "char",   # L1 cache control
                                "char",   # L3 cache control
                                "int",    # i32 address base
                                "anyint", # vNi32 address offsets
                                "short",  # address scale
                                "int",    # address immediate offset
                                "anyvector", # Data to write
                            ],
                            "attributes": "WriteMem", },
    "lsc_store_quad_ugm": { "result": "void",
                            "arguments": [
                                "anyint", # vNxi1, predicate
                                "char",   # address size
                                "char",   # element size
                                "char",   # channel mask
                                "char",   # L1 cache control
                                "char",   # L3 cache control
                                "long",   # i64 address base
                                "anyint", # vNi32 or vNi64 address offsets
                                "short",  # address scale
                                "int",    # address immediate offset
                                "anyvector", # Data to write
                            ],
                            "attributes": "WriteMem", },

### --------------------
### Thread ID intrinsics
### --------------------

## ``llvm.vc.internal.logical.thread.id`` : logical global thread ID
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
##
## * Return value: i32 logical global thread ID within a gpu tile
##
    "logical_thread_id" : { "result": "int",
                            "arguments": [],
                            "attributes": "NoMem", },

### ---------------------------
### Print and assert intrinsics
### ---------------------------

## ``llvm.vc.internal.assert.buffer`` : read stateless pointer to assert buffer
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
## ``llvm.vc.internal.assert.buffer`` : read implicit arg print assert ptr
##
## * return value: i64 address of assert buffer
##
    "assert_buffer" : { "result" : "long",
                        "arguments" : [],
                        "attributes" : "ReadMem", },

## ``llvm.vc.internal.print.buffer`` : read stateless pointer to print buffer
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
## ``llvm.vc.internal.print.buffer`` : read implicit arg print buffer ptr
##
## * return value: i64 address of print buffer
##
    "print_buffer" : { "result" : "long",
                       "arguments" : [],
                       "attributes" : "ReadMem", },

## ``llvm.vc.internal.print.format.index`` : add printf format string to collection
## ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
## ``llvm.vc.internal.print.format.index`` :  return index of printf format string
##
## * arg0: pointer for printf format string
##
## * Return value: the vector value read
##
    "print_format_index" : { "result" : "int",
                             "arguments" : ["anyptr"],
                             "attributes" : "NoMem", },
}



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

from typing import List, Set
from enum import Enum

class TypeID(Enum):
    Void = 0,
    Integer = 1,
    Float = 2,
    Vector = 3,
    Struct = 4,
    Pointer = 5,
    Any = 6,
    ArgumentReference = 7

    def __str__(self):
        return self.name

    @classmethod
    def from_str(cls, value : str):
        for key, val in cls.__members__.items():
            if key == value:
                return val
        else:
            raise ValueError("{value} is not present in {cls.__name__}")

class AddressSpace(Enum):
    Undefined = 0,
    Private = 1,
    Global = 2,
    Constant = 3,
    Local = 4,
    Generic = 5

    def __int__(self):
        return self.value[0] - 1

    @classmethod
    def from_str(cls, value : str):
        for key, val in cls.__members__.items():
            if key == value:
                return val
        else:
            raise ValueError("{value} is not present in {cls.__name__}")

class AttributeID(Enum):
    NoUnwind = 0,
    ReadNone = 1,
    ReadOnly = 2,
    ArgMemOnly = 3,
    WriteOnly = 4,
    NoReturn = 5,
    NoDuplicate = 6,
    Convergent = 7,
    InaccessibleMemOnly = 8

    def __str__(self):
        return self.name

    @classmethod
    def from_str(cls, value : str):
        for key, val in cls.__members__.items():
            if key == value:
                return val
        else:
            raise ValueError("{value} is not present in {cls.__name__}")

class TypeDefinition:
    def __init__(self, typeID : TypeID, bit_width : int  = 0, num_elements : int = 0,
                address_space : AddressSpace = AddressSpace.Undefined, internal_type = None,
                index : int = 0, internal_types = None):
        self.ID = typeID
        if self.ID == TypeID.Integer:
            self.bit_width = bit_width
        elif self.ID == TypeID.Float:
            self.bit_width = bit_width
        elif self.ID == TypeID.Any:
            self.default_type = internal_type
        elif self.ID == TypeID.Vector:
            self.num_elements = num_elements
            self.element_type = internal_type
        elif self.ID == TypeID.Pointer:
            self.address_space = address_space
            self.pointed_type = internal_type
        elif self.ID == TypeID.Struct:
            self.member_types = internal_types
        elif self.ID == TypeID.ArgumentReference:
            self.index = index

    def to_dict(self):
        if self.ID == TypeID.Integer:
            res = {
                "ID":  str(self.ID),
                "bit_width": self.bit_width
            }
        elif self.ID == TypeID.Float:
            res = {
                "ID":  str(self.ID),
                "bit_width": self.bit_width
            }
        elif self.ID == TypeID.Vector:
            res = {
                "ID":  str(self.ID),
                "num_elements": self.num_elements,
                "element_type": self.element_type.to_dict()
            }
        elif self.ID == TypeID.Pointer:
            res = {
                "ID":  str(self.ID),
                "address_space": str(self.address_space),
                "pointed_type": self.pointed_type.to_dict()
            }
        elif self.ID == TypeID.Struct:
            res = {
                "ID":  str(self.ID),
                "member_types": [ el.to_dict()for el in self.member_types ]
            }
        elif self.ID == TypeID.ArgumentReference:
            res = {
                "ID":  str(self.ID),
                "index": self.index
            }
        elif self.ID == TypeID.Void:
            res = {
                "ID":  str(self.ID)
            }
        elif self.ID == TypeID.Any:
            res = {
                "ID":  str(self.ID),
                "default_type": self.default_type.to_dict() if self.default_type != None else None
            }
        return res

    @staticmethod
    def from_dict(json_dct : dict):
        ID = TypeID.from_str(json_dct['ID'])
        if ID == TypeID.Integer:
            return TypeDefinition(ID, bit_width=json_dct['bit_width'])
        elif ID == TypeID.Float:
            return TypeDefinition(ID, bit_width=json_dct['bit_width'])
        elif ID == TypeID.Vector:
            internal_type = TypeDefinition.from_dict(json_dct['element_type'])
            return TypeDefinition(ID, num_elements=json_dct['num_elements'], internal_type=internal_type)
        elif ID == TypeID.Pointer:
            internal_type = TypeDefinition.from_dict(json_dct['pointed_type'])
            address_space = AddressSpace.from_str(json_dct['address_space'])
            return TypeDefinition(ID, address_space=address_space, internal_type=internal_type)
        elif ID == TypeID.Struct:
            member_types = [TypeDefinition.from_dict(el) for el in json_dct['member_types']]
            return TypeDefinition(ID, internal_types=member_types)
        elif ID == TypeID.ArgumentReference:
            return TypeDefinition(ID, index=json_dct['index'])
        elif ID == TypeID.Any:
            internal_type = TypeDefinition.from_dict(json_dct['default_type']) if json_dct['default_type'] else None
            return TypeDefinition(ID, internal_type=internal_type)

class ArgumentTypeDefinition:
    def __init__(self, type_definition : TypeDefinition, comment : str):
        self.type_definition = type_definition
        self.comment = comment

    def to_dict(self):
        res =  {
            "type_definition": self.type_definition.to_dict(),
            "comment": self.comment
        }
        return res

    @staticmethod
    def from_dict(json_dct : dict):
        type_definition = TypeDefinition.from_dict(json_dct['type_definition'])
        return ArgumentTypeDefinition(type_definition, json_dct['comment'])

class IntrinsicDefinition:
    def __init__(self, name : str, comment : str, return_type : ArgumentTypeDefinition,
                 argument_types : List[ArgumentTypeDefinition], attributes : Set[AttributeID]):
        self.name = name
        self.comment = comment
        self.return_type = return_type
        self.argument_types = argument_types
        self.attributes = sorted(list(attributes), key=lambda x: x.__str__())

    def to_dict(self):
        res = {
            "name": self.name,
            "comment": self.comment,
            "return_type": self.return_type.to_dict(),
            "argument_types":[ el.to_dict() for el in self.argument_types],
            "attributes": [str(el) for el in self.attributes]
        }
        return res

    @staticmethod
    def from_dict(json_dct : dict):
        return_type = ArgumentTypeDefinition.from_dict(json_dct['return_type'])
        argument_types = []
        for arg in json_dct['argument_types']:
            argument_types.append(ArgumentTypeDefinition.from_dict(arg))
        attributes = set(AttributeID.from_str(el) for el in json_dct['attributes'])
        return IntrinsicDefinition(json_dct['name'], json_dct['comment'], return_type,
                                   argument_types, attributes)


# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2017-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import binascii
import os
import re
import sys




def PrintHelp():
    sys.stdout.write('Usage: {0} <input file> <output file> <symbol name> <attribute>\n'.format(os.path.basename(__file__)))
    sys.stdout.write('\n')
    sys.stdout.write('    <input file>  - Path to input file which will be embedded.\n')
    sys.stdout.write('    <output file> - Path to output .cpp file which embedded data\n')
    sys.stdout.write('                    will be written to.\n')
    sys.stdout.write('    <symbol name> - Base name of symbol which identifies embedded data.\n')
    sys.stdout.write('    <attribute>   - "visibility" to add visibility attribute, "no_attr" to add no attribute\n')




class ResEmb:
    def __init__(self, inFile, outFile, symName, attrOpt):
        self.__symRe = re.compile('^[a-zA-Z_][a-zA-Z0-9_]*$')
        self.__lineSize = 20
        self.__chunkSize = 131072
        self.__symName = symName
        self.__attrOpt = attrOpt
        self.__openedFiles = list()
        self.__attr = ""
        self.__validate()
        try:
            self.__inFile = open(inFile, 'rb')
        # Note: open can throw OSError or IOError on Windows
        except EnvironmentError as ex:
            sys.stderr.write('ERROR: Cannot open input file "{0}".\n           {1}.\n'.format(inFile, ex.strerror))
            raise RuntimeError("failed to open input file")
        self.__openedFiles.append(self.__inFile)
        try:
            self.__outFile = open(outFile, 'w')
        except EnvironmentError as ex:
            sys.stderr.write('ERROR: Cannot create/open output file "{0}".\n       {1}.\n'.format(outFile, ex.strerror))
            raise RuntimeError("failed to open output file")
        self.__openedFiles.append(self.__outFile)

    def __del__(self):
        for openedFile in self.__openedFiles:
            openedFile.close()
    def __validate(self):
        symName = self.__symName
        if not self.__symRe.match(symName) or symName.endswith('_size'):
            sys.stderr.write('ERROR: Invalid symbol name "{0}".\n'.format(symName))
            raise ValueError

        attrOpt = self.__attrOpt
        if attrOpt == "visibility":
            self.__attr = '__attribute__((visibility("default")))'
        elif attrOpt == "no_attr":
            self.__attr = ""
        else:
            sys.stderr.write('ERROR: Invalid attribute argument: "{0}".\n'.format(attrOpt))
            raise ValueError

    def run(self):
        self.__outFile.write('// This file is auto generated by resource_embedder, DO NOT EDIT\n\n')
        self.__outFile.write('unsigned char {0} {1}[] = {{'.format(self.__attr, self.__symName))

        chunkSize = self.__chunkSize
        lineSize = self.__lineSize
        embeddedSize = 0;
        readBytes = self.__inFile.read(chunkSize)
        while len(readBytes) > 0:
            readSize = len(readBytes)
            hexBytes = binascii.hexlify(readBytes)

            if embeddedSize > 0:
                self.__outFile.write(',')
            self.__outFile.write(','.join((('\n        0x' if (embeddedSize + i) % lineSize == 0 else ' 0x') + hexBytes[2*i:2*i+2].decode("utf-8")) for i in range(readSize)))

            embeddedSize += readSize
            readBytes = self.__inFile.read(chunkSize)

        self.__outFile.write('\n    };\n\n');
        self.__outFile.write('unsigned int {0} {1}_size = {2};\n\n'.format(self.__attr, self.__symName, embeddedSize))


if __name__ == "__main__":
    numArgs = 5
    if len(sys.argv) < numArgs:
        PrintHelp()
        exit(0)
    for arg in sys.argv:
        if arg == '-h' or arg == '--help':
            PrintHelp()
            exit(0)

    if len(sys.argv) > numArgs:
        sys.stderr.write('WARNING: Number of arguments is greater than number of supported arguments.\n')
        sys.stderr.write('         All additional arguments will be ignored.\n')

    try:
        RE = ResEmb(sys.argv[1], sys.argv[2], sys.argv[3].strip(),
                sys.argv[4].strip())
        RE.run()

    except Exception as ex:
        sys.stderr.write('ERROR: Embeded script error.\n       {0}.\n'.format(repr(ex)))
        exit(1)




# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

# -*- Python -*-

import lit.formats
import lit.util

from lit.llvm import llvm_config
from lit.llvm.subst import ToolSubst
from lit.llvm.subst import FindTool

# Configuration file for the 'lit' test runner.

# name: The name of this test suite.
config.name = 'IGC'

# testFormat: The test format to use to interpret tests.
config.test_format = lit.formats.ShTest(not llvm_config.use_lit_shell)

# suffixes: A list of file extensions to treat as test files.
config.suffixes = ['.ll']

# excludes: A list of directories  and files to exclude from the testsuite.
config.excludes = ['CMakeLists.txt']

# test_source_root: The root path where tests are located.
config.test_source_root = os.path.dirname(__file__)

# test_exec_root: The root path where tests should be run.
config.test_exec_root = os.path.join(config.test_run_dir, 'test_output')

llvm_config.use_default_substitutions()

config.substitutions.append(('%PATH%', config.environment['PATH']))

tool_dirs = [config.igc_opt_dir, config.llvm_tools_dir]
tools = [ToolSubst('not'), ToolSubst('igc_opt')]

llvm_config.add_tool_substitutions(tools, tool_dirs)

# Add substitutions for pass configuration options to account for
# opt CLI changes between LLVM 10 and 11.
# FIXME: Remove once older-than-11 LLVM versions go out of use.
if int(config.llvm_version) < 11:
  config.substitutions.append(('%enable-basic-aa%', '-basicaa'))
else:
  config.substitutions.append(('%enable-basic-aa%', '--basic-aa'))
# Add LLVM version-dependent check prefixes.
# FIXME: Remove altogether after unifying all supported LLVM versions at 14+.
if int(config.llvm_version) < 14:
  config.substitutions.append(('%LLVM_DEPENDENT_CHECK_PREFIX%', 'CHECK-PRE-LLVM-14'))
else:
  config.substitutions.append(('%LLVM_DEPENDENT_CHECK_PREFIX%', 'CHECK-LLVM-14-PLUS'))
  config.available_features.add('llvm-14-plus')
if int(config.llvm_version) < 10:
  config.substitutions.append(('%LLVM_10_CHECK_PREFIX%', 'CHECK-PRE-LLVM-10'))
else:
  config.substitutions.append(('%LLVM_10_CHECK_PREFIX%', 'CHECK-LLVM-10-PLUS'))
  config.available_features.add('llvm-10-plus')

if config.use_khronos_spirv_translator_in_sc == "1":
  config.substitutions.append(('%SPV_CHECK_PREFIX%', 'CHECK-KHR'))
  config.available_features.add('khronos-translator')
else:
  config.substitutions.append(('%SPV_CHECK_PREFIX%', 'CHECK-LEGACY'))
  config.available_features.add('legacy-translator')
if not config.regkeys_disabled:
  config.available_features.add('regkeys')




#!/usr/bin/python3

# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2021-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import argparse
import json

class JSONWithCommentsDecoder(json.JSONDecoder):
  def __init__(self, **kw):
    super().__init__(**kw)

  def decode(self, s):
    s = '\n'.join(l for l in s.split('\n') if not l.lstrip(' ').startswith('//'))
    return super().decode(s)

OUTPUT_HEADER = """// AUTOGENERATED FILE, DO NOT EDIT!
// Generated by GenerateTranslationCode.py script."""
# C++ declarations separator.
INTERVAL_BETWEEN_DECLS = "\n\n"
BUILTIN_PREFIX = "__cm_cl_"

# The name of enum with builtin IDs.
BUILTIN_ID = "BuiltinID"
# The name of the enum with operand kinds and the suffix of builtin operand
# kind arrays.
OPERAND_KIND = "OperandKind"
# The suffix of builtin operand name enums.
OPERAND_NAME = "Operand"
# The name of the variable that holds a reference to builtin call.
BUILTIN_VARIABLE = "BiCall"
# The name of the variable that holds a reference to IR builder.
IRB_VARIABLE = "IRB"

# Section names.
BUILTIN_DESCS_SECTION = "CMCL_AUTOGEN_BUILTIN_DESCS"
TRANSLATION_DESCS_SECTION = "CMCL_AUTOGEN_TRANSLATION_DESCS"
TRANSLATION_IMPL_SECTION = "CMCL_AUTOGEN_TRANSLATION_IMPL"

parser = argparse.ArgumentParser(
  description="Generate translation code from JSON description.")
parser.add_argument("--desc", required=True,
  help="JSON file with a description", metavar="<input>.json")
parser.add_argument("--output", required=True, help="output file",
  metavar="<output>.inc")

# Opens \p desc_filename JSON file and parses it.
# Parsed structures are returned.
def get_description_from_json(desc_filename):
  with open(desc_filename, "r") as desc_file:
    return json.load(desc_file, cls=JSONWithCommentsDecoder)

# Generates:
# namespace name {
# enum Enum {
#   values[0],
#   values[1],
#   ...
# };
# } // namespace name
#
# The generated text is returned.
def generate_enum(name, values):
  text = "namespace {n} {{\nenum Enum {{\n".format(n=name)
  text += ",\n".join(["  {v}".format(v=value) for value in values])
  return text + "\n}};\n}} // namespace {n}".format(n=name)

# Generates:
# constexpr c_type name[] = {
#   values[0],
#   values[1],
#   ...
# };
#
# The generated text is returned.
def generate_array(c_type, name, values):
  assert values, "cannot generate an empty array"
  text = "constexpr {t} {n}[] = {{\n".format(t=c_type, n=name)
  text += ",\n".join(['  {v}'.format(v=value) for value in values])
  return text + "\n};"

# Generate enumerations that are not describing builtins but values of which
# are used to describe builtins.
def generate_helper_enums(helper_structures):
  return INTERVAL_BETWEEN_DECLS.join(
    [generate_enum(struct, helper_structures[struct])
      for struct in helper_structures])

def validate_builtin_desc(builtin_name, desc, helper_structures):
  if not all(operand["Kind"] in helper_structures[OPERAND_KIND]
               for operand in desc["Operands"]):
    raise RuntimeError("Some of {b} operand kinds is illegal because it's not "
                       "presented in OperandKind list".format(b=builtin_name))

# Raises an exception when some description inconsistency is found.
def validate_description(builtin_descs, helper_structures):
  for item in builtin_descs.items():
    validate_builtin_desc(*item, helper_structures)

# Returns a new list with additional "Size" element at the back.
def append_size(lst):
  return [*lst, "Size"]

# Generates an array with all the builtin names:
# costexpr const char* BuiltinNames[] = {
#   "__cm_cl_builtin0",
#   "__cm_cl_builtin1",
#   ...
# };
def generate_builtin_names_array(builtin_descs):
  return generate_array("const char*", "BuiltinNames",
                        ['"' + BUILTIN_PREFIX + desc["Name"] + '"'
                         for desc in builtin_descs.values()])

# Generates:
# namespace BuiltinOperand {
# enum Enum {
#   OperandName0,
#   OperandName1,
#   ...
# };
# } // namespace BuiltinOperand
def generate_operand_names_enum(builtin, desc):
  return generate_enum(
    builtin + OPERAND_NAME,
    append_size(operand["Name"] for operand in desc["Operands"]))

# Generates an enum for every builtin with its operands names to later use them
# as indices.
# Simplified output:
# enum Builtin0Operand { SRC };
# enum Builtin1Operand { DST, SRC };
# ...
def generate_operand_names_enums(builtin_descs):
  return INTERVAL_BETWEEN_DECLS.join(
    [generate_operand_names_enum(*builtin)
     for builtin in builtin_descs.items()])

# Generates an array with the number of operands for every builtin:
# constexpr int BuiltinOperandSize[] = {
#   Builtin0Operand::Size,
#   Builtin1Operand::Size,
#   ...
# };
def generate_operand_size_array(builtin_descs):
  return generate_array("int", "BuiltinOperandSize",
                        [builtin + OPERAND_NAME + "::Size"
                         for builtin in builtin_descs])

# Generates:
# constexpr OperandKind::Enum BuiltinOperandKind[] = {
#   OperandKind::Kind0,
#   OperandKind::Kind1,
#   ...
# };
def generate_operand_kinds_array(builtin, desc):
  return generate_array(OPERAND_KIND + "::Enum", builtin + OPERAND_KIND,
                        [OPERAND_KIND + "::" + operand["Kind"]
                         for operand in desc["Operands"]])

# Generates an array for every builtin with the list its operand kinds.
# Simplified output:
# constexpr OperandKind::Enum Builtin0OperandKind[] = {OperandKind::VectorIn};
# constexpr OperandKind::Enum Builtin1OperandKind[] = {
#   OperandKind::VectorOut, OperandKind::VectorIn};
def generate_operand_kinds_arrays(builtin_descs):
  return INTERVAL_BETWEEN_DECLS.join(
    generate_operand_kinds_array(builtin, desc)
      for builtin, desc in builtin_descs.items()
      if desc["Operands"])

# If there's an array of operand kinds, returns its name (array name degrades to
# pointer), otherwise returns nullptr. The can be operand kinds array if the
# builtin has no operands.
def get_operand_kinds_array_pointer(builtin, desc):
  if desc["Operands"]:
    return builtin + OPERAND_KIND
  return "nullptr"

# Generate an array of pointers to operand kinds arrays. So to get a kind of
# BuiltinN's M-th operand one can write BuiltinOperandKind[BuiltinN][M].
# Output:
# constexpr const OperandKind::Enum* BuiltinOperandKind[] = {
#   Builtin0OperandKind,
#   Builtin1OperandKind,
#   nullptr,
#   ...
# };
def generate_combined_operand_kinds_array(builtin_descs):
  return generate_array("const " + OPERAND_KIND + "::Enum*",
                        "Builtin" + OPERAND_KIND,
                        [get_operand_kinds_array_pointer(*builtin)
                         for builtin in builtin_descs.items()])

# Generate enums and arrays that describe CMCL builtins.
def generate_builtin_descriptions(builtin_descs):
  decls = [generate_enum(BUILTIN_ID, append_size(builtin_descs.keys())),
           generate_builtin_names_array(builtin_descs),
           generate_operand_names_enums(builtin_descs),
           generate_operand_size_array(builtin_descs),
           generate_operand_kinds_arrays(builtin_descs),
           generate_combined_operand_kinds_array(builtin_descs)]
  return INTERVAL_BETWEEN_DECLS.join(decls)

def begin_section(section_name):
  return "#ifdef " + section_name

def end_section(section_name):
  return "#endif // " + section_name

# Takes a list of section content and section name and returns a new list
# with section openning and closing strings at the begin and the end of the
# list.
def frame_section(section_content, section_name):
  return [begin_section(section_name),
          *section_content,
          end_section(section_name)]

# Generate a section of the output file that describes builtins.
def generate_builtin_descs_section(whole_desc):
  fragments = [generate_helper_enums(whole_desc["HelperStructures"]),
               generate_builtin_descriptions(whole_desc["BuiltinDescriptions"])]
  fragments = frame_section(fragments, BUILTIN_DESCS_SECTION)
  return INTERVAL_BETWEEN_DECLS.join(fragments)

# Generates an array of builtin handlers.
# Output:
# constexpr BuiltinCallHandler BuiltinCallHandlers[] = {
#   handleBuiltinCall<BuiltinID::Builtin0>,
#   handleBuiltinCall<BuiltinID::Builtin1>,
#   ...
# };
def generate_handlers_array(builtin_descs):
  return generate_array("BuiltinCallHandler", "BuiltinCallHandlers",
                        ["handleBuiltinCall<" + BUILTIN_ID + "::" +
                           builtin + ">"
                         for builtin in builtin_descs.keys()])

# Returns text representation for ID of intrinsic that was mentioned in
# "TranslateInto" section of builtin description. The section is taken as the
# argument. If there's no intrinsic, "~0u" is returned.
def get_intrinsic_id(translation_desc):
  if "VC-Intrinsic" in translation_desc:
    intrinsic = translation_desc["VC-Intrinsic"]
    if intrinsic.startswith("genx_"):
      return "GenXIntrinsic::" + intrinsic
    return "vc::InternalIntrinsic::" + intrinsic
  if "LLVM-Intrinsic" in translation_desc:
    return "Intrinsic::" + translation_desc["LLVM-Intrinsic"]
  return "~0u"

# Generates an array that represents map between builtin ID and intrinsic ID.
# Output:
# constexpr unsigned IntrinsicForBuiltin[] = {
#   IntrinsicForBuiltin0,
#   IntrinsicForBuiltin1,
#   ~0u, // Builtin2 has no corresponding intrinsic
#   ...
# };
def generate_intrinsics_array(builtin_descs):
  return generate_array("unsigned", "IntrinsicForBuiltin",
                        [get_intrinsic_id(desc["TranslateInto"])
                         for desc in builtin_descs.values()])

# Generate a section of the output file that holds some structures needed for
# the translation.
def generate_translation_descs_section(builtin_descs):
  fragments = [generate_handlers_array(builtin_descs),
               generate_intrinsics_array(builtin_descs)]
  fragments = frame_section(fragments, TRANSLATION_DESCS_SECTION)
  return INTERVAL_BETWEEN_DECLS.join(fragments)

# Generates code for GetBuiltinReturnType node.
# This node must have no arguments, so \p args is passed just to assert it.
def generate_builtin_return_type_expression(builtin_name, args):
  if args:
    raise RuntimeError("Builtin {bi} has invalid expession tree description: "
                       "GetBuiltinReturnType node must no arguments.".format(
                         bi=builtin_name))
  return "*{}.getType()".format(BUILTIN_VARIABLE)

# Get single operand from \p function_name node with required additional
# checks.
def get_single_operand_from_expression(function_name, builtin_name, args,
                                       builtin_desc):
  if len(args) != 1:
    raise RuntimeError("Builtin {bi} has invalid expession tree description: "
                       "{func} node must have only one argument.".format(
                         bi=builtin_name, func=function_name))
  operand = args[0]
  builtin_operands = [op["Name"] for op in builtin_desc["Operands"]]
  if not operand in builtin_operands:
    raise RuntimeError("Builtin {bi} has invalid expession tree description: "
                       "{func} argument is not an operand of this builtin."
                       .format(bi=builtin_name, func=function_name))
  return operand

# Generates code for GetBuiltinOperandType node.
# The node must have only one argument with the operand name.
# A call to getTypeFromBuiltinOperand is generated.
def generate_builtin_operand_type_expression(builtin_name, args, builtin_desc):
  operand = get_single_operand_from_expression("GetBuiltinOperandType",
                                               builtin_name,
                                               args,
                                               builtin_desc)
  return "getTypeFromBuiltinOperand<{bi_id}::{bi}>("\
         "{bi_call}, {bi}{op_suffix}::{op})".format(bi_id=BUILTIN_ID,
                                                    bi=builtin_name,
                                                    bi_call=BUILTIN_VARIABLE,
                                                    op_suffix=OPERAND_NAME,
                                                    op=operand)

# Generates a code that returns builtin operand value as llvm::Value& from
# GetBuiltinOperand node (or better to say leaf). The node must have a single
# argument which is a builtin operand name.
def generate_builtin_operand_expression(builtin_name, args, builtin_desc):
  operand = get_single_operand_from_expression("GetBuiltinOperand",
                                               builtin_name,
                                               args,
                                               builtin_desc)
  return "readValueFromBuiltinOp<{bi_id}::{bi}>({bi_call}, "\
         "{bi}{op_suffix}::{op}, {irb_var})".format(bi_id=BUILTIN_ID,
                                                    bi=builtin_name,
                                                    bi_call=BUILTIN_VARIABLE,
                                                    op_suffix=OPERAND_NAME,
                                                    op=operand,
                                                    irb_var=IRB_VARIABLE)

# Generate code for Code node. If Code node has a single argument, it is a
# string and it is the generated code. If the node has multiple arguments, the
# first argument is a python format string, the rest of arguments are the
# arguments for this string. Arguments must be expression trees too.
def generate_code_expression(builtin_name, args, builtin_desc):
  if not args:
    raise RuntimeError("Builtin {bi} has invalid expession tree description: "
                       "Code node must have at least one argument.".format(
                         bi=builtin_name))
  code = args[0]
  if not isinstance(code, str):
    raise RuntimeError("Builtin {bi} has invalid expession tree description: "
                       "Code node must have a string as the first argument."
                       .format(bi=builtin_name))
  if len(args) == 1:
    return code
  replacements = [generate_expression_tree(builtin_name, arg, builtin_desc)
                  for arg in args[1:]]
  code = code.format(*replacements)
  return code

# Generate a complex nested expression based on description in \p desc_tree.
# Each tree node has the following structure: {name: [args]} where name is the
# name of the node, args - the list of node aruments. Args may contain nodes,
# thus the whole structure is a tree.
# \p builtin_desc is passed for validation.
def generate_expression_tree(builtin_name, tree_desc, builtin_desc):
  if len(tree_desc) != 1:
    raise RuntimeError("Builtin {bi} has invalid expession tree description: "
                       "Object must have only one entry.".format(
                         bi=builtin_name))
  function = next(iter(tree_desc))
  args = tree_desc[function]
  if not isinstance(args, list):
    raise RuntimeError("Builtin {bi} has invalid expession tree description: "
                       "Object entry item must be an Array.".format(
                         bi=builtin_name))
  if function == "GetBuiltinReturnType":
    return generate_builtin_return_type_expression(builtin_name, args)
  if function == "GetBuiltinOperandType":
    return generate_builtin_operand_type_expression(builtin_name, args,
                                                    builtin_desc)
  if function == "GetBuiltinOperand":
    return generate_builtin_operand_expression(builtin_name, args,
                                               builtin_desc)
  if function == "Code":
    return generate_code_expression(builtin_name, args, builtin_desc)
  raise RuntimeError("Builtin {bi} has invalid expession tree description: "
                     "Unknown node.".format(bi=builtin_name))

# Output:
# template <>
# Type &getTranslatedBuiltinType<BuiltinID::Builtin>(CallInst &BiCall) {
#   return .....;
# };
def generate_return_type_specialization(builtin_name, desc):
  text = "template <>\n"
  text += "Type &getTranslatedBuiltinType<{bi_enum}::{bi}>(CallInst &"\
          "{bi_var}) {{\n".format(bi_enum=BUILTIN_ID, bi=builtin_name,
                                  bi_var=BUILTIN_VARIABLE)
  text += "  return {};\n".format(
      generate_expression_tree(builtin_name,
                               desc["TranslateInto"]["ReturnType"],
                               desc))
  text += "}"
  return text

# Generates getTranslatedBuiltinOperands specialization for builtin with the
# name \p builtin_name. The rule for every operand generation is defined as
# expression tree in TranslateInto:Operands section of JSON description.
#
# Output:
# template <>
# std::vector<Value *>
# getTranslatedBuiltinOperands<BuiltinID::Builtin>(CallInst &BiCall,
#                                                  IRBuilder<> &IRB) {
#   return {.....};
# }
def generate_operand_specialization(builtin_name, desc):
  text = "template <>\n"
  text += "std::vector<Value *>\n"
  text += "getTranslatedBuiltinOperands<{bi_enum}::{bi}>(CallInst &{bi_var}, "\
          "IRBuilder<> &{irb_var}) {{\n".format(bi_enum=BUILTIN_ID,
                                                bi=builtin_name,
                                                bi_var=BUILTIN_VARIABLE,
                                                irb_var=IRB_VARIABLE)
  operands = [generate_expression_tree(builtin_name, op_desc, desc)
              for op_desc in desc["TranslateInto"]["Operands"]]
  operands = ["&" + op for op in operands]
  operands = ",\n          ".join(operands)
  text += "  return {{{}}};\n".format(operands)
  text += "}"
  return text

# Generates getTranslatedBuiltinType specialization for every builtin.
def generate_return_type_function(builtin_descs):
  return INTERVAL_BETWEEN_DECLS.join(
      [generate_return_type_specialization(*item)
       for item in builtin_descs.items()])

# Generates getTranslatedBuiltinOperands specialization for every builtin.
def generate_operand_function(builtin_descs):
  return INTERVAL_BETWEEN_DECLS.join(
      [generate_operand_specialization(*item)
       for item in builtin_descs.items()])

# Generates a section that will hold some implementation required for
# builtin translation.
def generate_translation_impl_section(builtin_descs):
  fragments = [generate_return_type_function(builtin_descs),
               generate_operand_function(builtin_descs)]
  fragments = frame_section(fragments, TRANSLATION_IMPL_SECTION)
  return INTERVAL_BETWEEN_DECLS.join(fragments)

# Generate output file text.
def generate_file(whole_desc):
  validate_description(whole_desc["BuiltinDescriptions"],
                       whole_desc["HelperStructures"])
  sections = [OUTPUT_HEADER,
              generate_builtin_descs_section(whole_desc),
              generate_translation_descs_section(whole_desc["BuiltinDescriptions"]),
              generate_translation_impl_section(whole_desc["BuiltinDescriptions"])]
  return INTERVAL_BETWEEN_DECLS.join(sections)

args = parser.parse_args()
whole_desc = get_description_from_json(args.desc)
output_str = generate_file(whole_desc)
with open(args.output, "w") as output_file:
  output_file.write(output_str)



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2017-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import os
import sys
import errno
import re
from typing import List, Tuple, TextIO, Callable, Generator

class DeclHeader:
    # line contains the entire string of the line the decl was found on
    line: str
    # declName is just the identifier name
    declName: str
    # fields contains a list of all the names of the fields in the structure
    fields: List[str]

    def __init__(self, line: str, declName: str, fields: List[str]):
        self.line     = line
        self.declName = declName
        self.fields   = fields

enumNames: List[DeclHeader] = []
structureNames: List[DeclHeader] = []

def parseCmdArgs() -> Tuple[str, str]:
    if (len(sys.argv) != 3):
        sys.exit("usage: autogen.py <path_to_MDFrameWork.h> <path_to_MDNodeFuncs.gen>")

    __MDFrameWorkFile__ = sys.argv[1]
    __genFile__         = sys.argv[2]

    if not os.path.isfile(__MDFrameWorkFile__):
        sys.exit(f"Could not find the file {__MDFrameWorkFile__}")

    __genDir__ = os.path.dirname(__genFile__)
    if not os.path.exists(__genDir__):
        try:
            os.makedirs(__genDir__)
        except OSError as err:
            if err.errno != errno.EEXIST:
                sys.exit(f"Failed to create the directory {__genDir__}")

    return __MDFrameWorkFile__ , __genFile__

def extractStructField(line: str, declHeader: DeclHeader):
    if line.strip() == '':
        return
    vars = line.split()
    if "=" in line:
        declHeader.fields.append(vars[vars.index("=") - 1] + ";")
    else:
        declHeader.fields.append(vars[-1])

def extractEnumVal(line: str, declHeader: DeclHeader):
    vars = line.split()
    if len(vars) == 0 or "{" in line:
        return

    val = vars[0]
    if val[-1] == ',':
        val = val[:-1]

    declHeader.fields.append(val)

def lines(s: str) -> Generator[str, None, None]:
    for line in s.split('\n'):
        yield line

def parseHeader(fileContents: str):
    insideIGCNameSpace = False
    pcount = 0
    file = lines(fileContents)
    for line in file:
        line = line.split("//")[0]
        if "namespace IGC" in line:
            while "{" not in line:
                line = next(file, None)
                if line is None:
                    sys.exit('missing opening brace!')
            insideIGCNameSpace = True
            pcount += 1
        if insideIGCNameSpace:
            blockType = re.search("struct|enum", line)
            if blockType:
                words = line.split()
                idx = 2 if 'class' in words else 1
                foundDecl = DeclHeader(line, words[idx], [])
                opcount = pcount
                namesList = structureNames
                extractFunc = extractStructField
                if blockType[0] == 'enum':
                    namesList = enumNames
                    extractFunc = extractEnumVal
                while True:
                    line = next(file, None)
                    if line is None:
                        sys.exit(f"EOF reached with unclosed enum or struct, check formatting")
                    line = line.split("//")[0]
                    pcount += line.count("{") - line.count("}")
                    if pcount <= opcount:
                        break
                    extractFunc(re.sub("{|}","", line), foundDecl)
                assert pcount == opcount, f"Unexpected struct/enum ending, check formatting"
                namesList.append(foundDecl)
            elif "}" in line and "};" not in line:
                insideIGCNameSpace = False
                pcount -= 1
    assert pcount == 0, f"EOF reached, with unclosed IGC namespace, check formatting"

def stripBlockComments(text: str) -> str:
    return re.sub(r'/\*(.|\s)*?\*/', '', text)

def expandIncludes(fileName: str) -> str:
    try:
        file = open(fileName, 'r')
    except:
        sys.exit(f"Failed to open the file {fileName}")

    text = file.read()
    while True:
        # look for includes of the form: #include "myinclude.h" // ^MDFramework^
        includes: List[Tuple[str, str]] = []
        for m in re.finditer(r'#include\s+"(\S+)"\s*//\s*\^MDFramework\^:\s*(\S+)', text):
            include_file = os.path.basename(m.group(1))
            relative_path = m.group(2)
            parent_dir = os.path.dirname(fileName)
            include_file_path = os.path.normpath(
                os.path.join(parent_dir, relative_path, include_file))
            includes.append((m.group(0), include_file_path))

        if len(includes) == 0:
            break

        for (include_string, include_path) in includes:
            try:
                file = open(include_path, 'r')
            except:
                sys.exit(f"Failed to open the file {include_path}")
            include_contents = file.read()
            text = text.replace(include_string, include_contents)

    return text

def printStructCalls(structDecl: DeclHeader, outputFile: TextIO):
    outputFile.write("    Metadata* v[] = \n")
    outputFile.write("    { \n")
    outputFile.write("        MDString::get(module->getContext(), name),\n")
    for item in structDecl.fields:
        item = item[:-1]
        outputFile.write(f"        CreateNode({structDecl.declName}Var.{item}, module, ")
        outputFile.write(f'"{item}"')
        outputFile.write("),\n")
    outputFile.write("    };\n")
    outputFile.write("    MDNode* node = MDNode::get(module->getContext(), v);\n")
    outputFile.write("    return node;\n")

def printEnumCalls(enumDecl: DeclHeader, outputFile: TextIO):
    outputFile.write("    StringRef enumName;\n")
    outputFile.write(f"    switch({enumDecl.declName}Var)\n")
    outputFile.write("    {\n")
    for item in enumDecl.fields:
        outputFile.write(f"        case IGC::{enumDecl.declName}::{item}:\n")
        outputFile.write("            enumName = ")
        outputFile.write(f'"{item}"')
        outputFile.write(";\n")
        outputFile.write("            break;\n" )
    outputFile.write("    }\n")
    outputFile.write("    Metadata* v[] = \n")
    outputFile.write("    { \n")
    outputFile.write("        MDString::get(module->getContext(), name),\n")
    outputFile.write("        MDString::get(module->getContext(), enumName),\n")
    outputFile.write("    };\n")
    outputFile.write("    MDNode* node = MDNode::get(module->getContext(), v);\n")
    outputFile.write("    return node;\n")

def printStructReadCalls(structDecl: DeclHeader, outputFile: TextIO):
     for item in structDecl.fields:
        item = item[:-1]
        outputFile.write(f"    readNode({structDecl.declName}Var.{item}, node , ")
        outputFile.write(f'"{item}"')
        outputFile.write(");\n")

def printEnumReadCalls(enumDecl: DeclHeader, outputFile: TextIO):
    outputFile.write("    StringRef s = cast<MDString>(node->getOperand(1))->getString();\n")
    outputFile.write("    std::string str = s.str();\n")
    outputFile.write(f"    {enumDecl.declName}Var = (IGC::{enumDecl.declName})(0);\n")

    for item in enumDecl.fields:
        outputFile.write(f'    if((str.size() == sizeof("{item}")-1) && (::memcmp(str.c_str(),')
        outputFile.write(f'"{item}"')
        outputFile.write(",str.size())==0))\n")
        outputFile.write("    {\n")
        outputFile.write(f"            {enumDecl.declName}Var = IGC::{enumDecl.declName}::{item};\n")
        outputFile.write("    } else\n")

    outputFile.write("    {\n")
    outputFile.write(f"            {enumDecl.declName}Var = (IGC::{enumDecl.declName})(0);\n")
    outputFile.write("    }\n")

def emitCodeBlock(names: List[DeclHeader], fmtFn: Callable[[str], str], printFn: Callable[[DeclHeader, TextIO], None], outputFile: TextIO):
    for item in names:
        outputFile.write(fmtFn(item.declName))
        outputFile.write("{\n")
        printFn(item, outputFile)
        outputFile.write("}\n\n")

def emitEnumCreateNode(outputFile: TextIO):
    def fmtFn(item: str):
        return f"MDNode* CreateNode(IGC::{item} {item}Var, Module* module, StringRef name)\n"
    emitCodeBlock(enumNames, fmtFn, printEnumCalls, outputFile)

def emitStructCreateNode(outputFile: TextIO):
    def fmtFn(item: str):
        return f"MDNode* CreateNode(const IGC::{item}& {item}Var, Module* module, StringRef name)\n"
    emitCodeBlock(structureNames, fmtFn, printStructCalls, outputFile)

def emitEnumReadNode(outputFile: TextIO):
    def fmtFn(item: str):
        return f"void readNode( IGC::{item} &{item}Var, MDNode* node)\n"
    emitCodeBlock(enumNames, fmtFn, printEnumReadCalls, outputFile)

def emitStructReadNode(outputFile: TextIO):
    def fmtFn(item: str):
        return f"void readNode( IGC::{item} &{item}Var, MDNode* node)\n"
    emitCodeBlock(structureNames, fmtFn, printStructReadCalls, outputFile)

def genCode(fileName: str):
    try:
        outputFile = open(fileName, 'w')
    except:
        sys.exit(f"Failed to open the file {fileName}")

    emitEnumCreateNode(outputFile)
    emitStructCreateNode(outputFile)
    emitEnumReadNode(outputFile)
    emitStructReadNode(outputFile)

if __name__ == '__main__':
    __MDFrameWorkFile__ , __genFile__ = parseCmdArgs()
    expansion = expandIncludes(__MDFrameWorkFile__)
    expansion = stripBlockComments(expansion)
    parseHeader(expansion)
    genCode(__genFile__)



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2017-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

# import argparse
# import os
# def concat(file1,file2,new_file):
#     f1,f2 = "",""
#     if os.path.exists(file1):
#         f1 = open(file1).read()
#     if os.path.exists(file2):
#         f2 = open(file2).read()
#     with open(new_file,"wb") as f_concat:
#         f_concat.write(f1+f2)
#         f_concat.write("\0")
# parser = argparse.ArgumentParser()
# parser.add_argument('-top', help='top file for concat')
# parser.add_argument('-bottom', help='bottom file for concat')
# parser.add_argument('-new', help='new file for concat')
# args = parser.parse_args()
# concat(args.top,args.bottom,args.new)

import os
import sys

def Usage () :
  print (r'Usage: concat.py -new <outfile> <f0> <f1> ...')
  print (r'  To concatenate files in order into a new file <outfile>.')
  print (r'  Input files are <f0> <f1>, and they are at least two.')

if len(sys.argv) < 5 :
  Usage()
  sys.exit("Error: not enough argments to concat.py")

if sys.argv[1] != '-new':
  Usage()
  sys.exit("Error: concat.py - the 1st argument must be -new <outfile>")

outfilename = sys.argv[2]
infilenames = sys.argv[3:]

with open(outfilename, 'wb') as outfile:
  for fname in infilenames:
    if not os.path.exists(fname):
      sys.exit("Error: no such file: " + fname)
    with open(fname, "rb") as infile:
      outfile.write(infile.read())
  outfile.write(b"\0")



#!/usr/bin/python3

# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2019-2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

"""
Usage: cisa_gen_intrinsics.py <input_file> <output_path>

This script gets intrinsics description from JSON file specified by <input_file> argument
and generates two files GenXIntrinsicInfoTable.inc and GenXIntrinsicsBuildMap.inc into
path specified by <output_path> argument.

JSON file must contain following mandatory fields: INTRINSICS, OPCODE_GEN and ARGUMENTS_GEN.

*** Field INTRINSICS
  Contains description of all intrinsics. Each intrinsic is described in following format:
  intrinsic_name : {
    opc: VISA opcode corresponding to the intrinsic
    gen_opc: optional field, it aims to distinguish generators of complex opcodes which may
             contain sub-opcode field
    OPTIONS: list of intrinsics options. Currently, supported only 'disable' value, which means
             that intrinsic will be skipped at all.
    <ARGUMENT>: see description below
  }

  Each argument is a [key: list] format, where key is a name of Argument, list is a command
  for generator.
  First field of generator command is a generator name, it tells how to generate code for
  fetching an argument value. Each argument generator is described in ARGUMENTS_GEN map.

  For example:
      "Surface": ["GENERAL", "UNSIGNED", 10],
  Here GENERAL is generator name by which will be determined (from "ARGUMENTS_GEN") what code
  to generate for getting argument value.
  Generated code:
      auto Surface = CreateOperand(II::ArgInfo(UNSIGNED | 10));
  or for GenXIntrinsicInfoTable.inc:
      GENERAL | UNSIGNED | 10,

  To add new intrinsic you need to add new description into INTRINSICS map. If it contains
  opcode which is absent in opcode_map you also need to add item for new opcode to OPCODE_GEN.

  For example, lets add new intrinsic with new opcode and one new argument generator(NEW_PREDICATION):
  "INTRINSICS":
    "genx_new": {
      "opc": "ISA_NEW",
      "exec_size": ["EXECSIZE_FROM_ARG", 1],
      "pred": ["NEW_PREDICATION", 1],
      "DataOrder": ["BYTE", 5],
      "Surface": ["GENERAL", "UNSIGNED", 10],
      "DstData": ["RAW", 0],
      "Src1Data": ["NULLRAW"]
    },
  "OPCODE_GEN":
    ISA_NEW: "CISA_CALL(Kernel->AppendNew(exec_size, pred, DataOrder, Src1Data, DstData, Surface));"
  "ARGUMENTS_GEN":
    "NEW_PREDICATION": "CreateNewPredication(II::ArgInfo({args}))",
  Also we need to add new function or lambda with name CreateNewPredication to GenXCisaBuilder.cpp

*** Field ARGUMENTS_GEN
  It is needed only to generate CISA building code (GenXIntrinsicsBuildMap.inc)
  Pattern keys that can be used inside generator:
  args   - string with arguments that are passed to ArgInfo constructor.
  value1 - first value in argument list, needed for LITERAL generator
  dst    - name of a variable to which will be assigned argument value

*** Field OPCODE_GEN
  It is needed only to generate CISA building code (GenXIntrinsicsBuildMap.inc)
  Final part of generated code for a single intrinsic is a call of Finalizer's function that builds
  instruction itself. So, all items of this map is just map from opcode to the build function.
  Opcode may be not real VISA opcode, for example, ISA_VA_SKL_PLUS has different functions to build
  instructions with different signatures, which depends of its sub-opcode. Thus there are maybe
  compound opcodes for such cases.
"""

import sys
import re
import json
from collections import OrderedDict


HEADER = '''/******************************************************************************
 * AUTOGENERATED FILE, DO NOT EDIT!
 * Generated by GenXUtilBuild project
 */
'''

def open_and_delete_comments(dscr_filename):
    with open(dscr_filename, "r") as jsonfile:
        data = jsonfile.readlines()
    jsonwithoutcomments = filter(lambda line: "//" not in line, data)
    stringjson = "".join(jsonwithoutcomments)
    return stringjson;

def generate(dscr_filename, out_path):
    special_keys = ('gen_opc', 'OPTIONS')
    descr = json.loads(open_and_delete_comments(dscr_filename), object_pairs_hook=OrderedDict)
    opcode_gen = descr['OPCODE_GEN']
    arguments_gen = descr['ARGUMENTS_GEN']
    intrinsics = descr['INTRINSICS']

    # Convert list to function call string
    # Example: [ Func, arg1, arg2] to Func(arg1, arg2)
    def gen2str(value):
        if isinstance(value, list):
            args = []
            for v in value[1:]:
                args.append(gen2str(v))
            return "{}({})".format(value[0], ', '.join(args))
        return str(value)

    # Recursively search regex in lists
    def gen_search(value, regex):
        if isinstance(value, list):
            for v in value:
                if gen_search(v, regex):
                    return True
            return False
        return bool(re.search(regex, value))

    def isstrinst(opc_gen):
        isalt = True
        if sys.version_info[0] >= 3:
            isalt = isinstance(opc_gen, bytes)
        else:
            isalt = isinstance(opc_gen, unicode)
        return bool(isinstance(opc_gen, str) or isalt)

    def getIntrinsicName(name):
        if name.startswith('genx_'):
            return f'GenXIntrinsic::{name}'
        if name.startswith('vc::InternalIntrinsic'):
            return name
        return f'Intrinsic::{name}'

    def getIntrinsicInfoStr(name, intr):
        intr = list(filter(lambda arg: arg[0] not in special_keys, intr.items()))
        args = []
        for key, value in intr:
            if key in ('opc'):
                args.append('LITERAL | {}'.format(value))
            elif isinstance(value, list):
                should_skip_operand = lambda x: x == 'RAW_OPERANDS' or ('BUILD_ONLY::') in str(x)
                args.append(' | '.join([str(x) for x in value if not should_skip_operand(x)]))

        return '{{ {}, {{ {} }} }}'.format(getIntrinsicName(name), ', '.join(args))

    should_skip_intrinsic = lambda intr: 'OPTIONS' in intr and 'disable' in intr['OPTIONS']
    intrinsics = list(filter(lambda intr: not should_skip_intrinsic(intr[1]), intrinsics.items()))

    with open(out_path + '/GenXIntrinsicInfoTable.inc', 'w') as file:
        file.write(HEADER)
        file.write(',\n'.join([getIntrinsicInfoStr(*intr) for intr in intrinsics]))

    def analyseForBuildMap(x):
        if isstrinst(x) and 'BUILD_ONLY::' not in str(x):
            return 'II::' + x
        elif 'BUILD_ONLY::' in str(x):
            return str(x).rsplit('BUILD_ONLY::',1)[1]
        else:
            return str(x)

    with open(out_path + '/GenXIntrinsicsBuildMap.inc', 'w') as file:
        file.write(HEADER)
        file.write('switch(IntrinID) {\n\n')

        for name, intr in intrinsics:
            gen_opc = intr.get('gen_opc')
            if not gen_opc:
                gen_opc = intr['opc']

            opc_gen = opcode_gen.get(gen_opc)
            if not opc_gen:
                print(intr)
                raise RuntimeError("Instruction generator not found")
            if isstrinst(opc_gen):
                opc_gen = [opc_gen]
            assert isinstance(opc_gen, list)

            file.write('  case {}: {{\n'.format(getIntrinsicName(name)))

            for key, value in intr.items():
                if key in special_keys:
                    continue

                # no_assign means that there is no variable that need to be assigned
                no_assign = key in ('twoaddr', 'nobarrier')

                # skip items that are not exist in generator string
                if not no_assign and not gen_search(opc_gen, r'\b%s\b'%key):
                    continue

                if key == 'opc':
                    replace = value
                elif isinstance(value, list):
                    replace = arguments_gen.get(value[0])
                    if not replace:
                        print(value)
                        raise RuntimeError('Key not found!')
                    if not replace:
                        continue
                    context = { 'value1': value[1] if len(value) > 1 else None, 'dst': key,
                                'args': '{}'.format(' | ').join(
                                [analyseForBuildMap(x) for x in value if x != 'RAW_OPERANDS']) }
                    if isinstance(replace, list):
                        replace = [x.format(**context) for x in replace]
                    else:
                        replace = replace.format(**context)
                else:
                    replace = value
                assert replace, 'Unknown token'

                if isinstance(replace, list):
                    for replace_item in replace:
                        file.write('    ' + replace_item + ';\n')
                else:
                    assign = '' if no_assign else 'auto ' + key + ' = '
                    file.write('    ' + assign + replace + ';\n')

            for g in opc_gen:
                file.write('    ' + gen2str(g) + ';\n')
            file.write('  } break;\n\n')

        file.write('''  default:
    CI->print(errs());
    errs() << '\\n';
    report_fatal_error("Unsupported intrinsic!");
    break;
}''')

def main():
    if len(sys.argv) > 1 and sys.argv[1] == '--help':
        print(__doc__)
        sys.exit(0)
    assert len(sys.argv) > 2, "Missing arguments! Usage: cisa_gen_intrinsics.py <INPUT_FILE> <OUTPUT_PATH>"
    generate(sys.argv[1], sys.argv[2])

if __name__ == '__main__':
    main()



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import sys
import json
import argparse
from typing import List
from Intrinsic_definition_objects import *
from Intrinsic_utils import *
from itertools import takewhile

class IntrinsicLookupTableEntry:
    def __init__(self, id : str, lookup_name : str, common_prefix_len : int):
        self.id = id
        self.lookup_name = lookup_name
        self.common_prefix_len = common_prefix_len

class IntrinsicFormatter:
    use_comments = False

    @staticmethod
    def get_lookup_table(intrinsic_definitions : List[IntrinsicDefinition]):
        lookup_table = []
        for el in intrinsic_definitions:
            lookup_table.append(IntrinsicLookupTableEntry(el.name, IntrinsicFormatter.get_lookup_name(el.name), 0))
        lookup_table.sort(key=lambda x: x.lookup_name[:-1])
        for i in range(1, len(lookup_table)):
            prev = lookup_table[i - 1].lookup_name
            curr = lookup_table[i].lookup_name
            lookup_table[i].common_prefix_len =  sum(1 for _ in takewhile(lambda x: x[0] == x[1], zip(prev, curr)))
            lookup_table[i - 1].common_prefix_len = max(
                lookup_table[i - 1].common_prefix_len,
                lookup_table[i].common_prefix_len)
        return lookup_table

    def get_intrinsic_lookup_table_entry_initialization_list(lookup_table_entry : IntrinsicLookupTableEntry, is_last):
        lookup_name = lookup_table_entry.lookup_name
        output = '{{ llvm::GenISAIntrinsic::ID::{}, {}, "{}" }}'.format(
            lookup_table_entry.id,
            lookup_table_entry.common_prefix_len,
            lookup_table_entry.lookup_name)
        if not is_last:
            output = "{},".format(output)
        return output

    @staticmethod
    def get_argument_name(argument_type, index):
        output = "Arg{}".format(index)
        if hasattr(argument_type, 'name'):
            output = argument_type.name
        if (index == 0):
            output = "{} = 0".format(output)
        output = "{},".format(output)
        return output

    @staticmethod
    def get_intrinsic_suffix(name):
        name = name[len("GenISA_"):]
        output = name.replace("_", ".")
        return output

    @staticmethod
    def format_name(name):
        output = '"{}"'.format(name)
        return output

    @staticmethod
    def get_lookup_name(name):
        name = name[len("GenISA_"):] + '@'
        return name

    @staticmethod
    def get_type_definition(type_def):
        output = "Unknown"
        if type_def == None:
            output = "EmptyTypeHolderT"
        elif type_def.ID == TypeID.Integer:
            output = "IntegerTypeHolderT<{}>".format(type_def.bit_width)
        elif type_def.ID == TypeID.Float:
            output = "FloatTypeHolderT<{}>".format(type_def.bit_width)
        elif type_def.ID == TypeID.Vector:
            assert(type_def.element_type)
            element_type_name = IntrinsicFormatter.get_type_definition(type_def.element_type)
            if type_def.num_elements > 0:
                output = "VectorTypeHolderT<{}, {}>".format(
                    element_type_name,
                    type_def.num_elements)
            else:
                output = "VectorTypeHolderT<{}>".format(
                    element_type_name)
        elif type_def.ID == TypeID.Pointer:
            assert(type_def.pointed_type)
            assert(type_def.address_space)
            pointed_type_name = IntrinsicFormatter.get_type_definition(type_def.pointed_type)
            address_space = int(type_def.address_space)
            if address_space >= 0:
                output = "PointerTypeHolderT<{}, {}>".format(
                    pointed_type_name,
                    address_space)
            else:
                output = "PointerTypeHolderT<{}>".format(
                    pointed_type_name)
        elif type_def.ID == TypeID.Struct:
            output = "StructTypeHolderT<MemberTypeListHolderT<{}>>".format(
                ", ".join([ "{}".format(IntrinsicFormatter.get_type_definition(member_type)) for member_type in type_def.member_types ]))
        elif type_def.ID == TypeID.ArgumentReference:
            output = "ArgumentReferenceTypeHolderT<{}>".format(type_def.index)
        elif type_def.ID == TypeID.Void:
            output = "EmptyTypeHolderT"
        elif type_def.ID == TypeID.Any:
            if type_def.default_type:
                default_type_name = IntrinsicFormatter.get_type_definition(type_def.default_type)
                output = "AnyTypeHolderT<{}>".format(default_type_name)
            else:
                output = "AnyTypeHolderT<>"
        return output

    @staticmethod
    def get_argument_type_entry(type_def, is_last):
        output = IntrinsicFormatter.get_type_definition(type_def)
        output = "{}::scType".format(output)
        if not is_last:
            output = "{},".format(output)
        return output

    @classmethod
    def get_comment(cls ,comment):
        if cls.use_comments:
            output = '"{}"'.format(comment)
            return output
        return '""'

    @staticmethod
    def get_argument_comment(comment, is_last):
        output = IntrinsicFormatter.get_comment(comment)
        if not is_last:
            output = "{},".format(output)
        return output

    @staticmethod
    def get_attribute_entry(attribute, is_last):
        output = 'llvm::Attribute::{}'.format(attribute)
        if not is_last:
            output = "{},".format(output)
        return output

    @staticmethod
    def get_prefix():
        prefix = 'llvm.genx.GenISA.'
        return prefix

def generate_intrinsic_defintion_files(intrinsic_definitions : List[IntrinsicDefinition], output_directory : str, use_comments):

    IntrinsicFormatter.use_comments = use_comments

    template_lookup = TemplateLookup(directories=[r'.'])
    template = Template(filename=r'templates/GenIntrinsicEnum.h.mako',
                        lookup=template_lookup)
    output_file_path = os.path.join(
        output_directory, from_template_name_to_destination_name(template.filename))
    write_to_file_using_template(output_file_path, template, intrinsic_definitions=intrinsic_definitions)

    template = Template(filename=r'templates/GenIntrinsicDescription.h.mako',
                        lookup=template_lookup)
    output_file_path = os.path.join(
        output_directory, from_template_name_to_destination_name(template.filename))
    write_to_file_using_template(output_file_path, template, intrinsic_definitions=intrinsic_definitions)

    template = Template(filename=r'templates/GenIntrinsicDefinition.h.mako',
                        lookup=template_lookup)
    output_file_path = os.path.join(
        output_directory, from_template_name_to_destination_name(template.filename))
    write_to_file_using_template(output_file_path, template, intrinsic_definitions=intrinsic_definitions)

    template = Template(filename=r'templates/GenIntrinsicDefinition.cpp.mako',
                        lookup=template_lookup)
    output_file_path = os.path.join(
        output_directory, from_template_name_to_destination_name(template.filename))
    write_to_file_using_template(output_file_path, template, intrinsic_definitions=intrinsic_definitions)

    template = Template(filename=r'templates/GenIntrinsicLookupTable.h.mako',
                        lookup=template_lookup)
    output_file_path = os.path.join(
        output_directory, from_template_name_to_destination_name(template.filename))
    write_to_file_using_template(output_file_path, template, intrinsic_definitions=intrinsic_definitions)

if __name__ == '__main__':
    def main(args):
        parser = argparse.ArgumentParser(description='Generates IGC intrinsic files.')
        parser.add_argument("inputs", nargs='+', help="the source path to the file with intrinsic defintions (current IGC format)",
                        type=file_path)
        parser.add_argument("--output", help="the directory for the files with intrinsic definitions",
                        type=dir_path)
        parser.add_argument("--use_comments", action='store_true')

        args = parser.parse_args(args)
        raw_data = []
        for el in args.inputs:
            with open(el) as f:
                raw_data.extend(json.load(f))

        intrinsic_definitions = [ IntrinsicDefinition.from_dict(el) for el in raw_data ]

        if len(intrinsic_definitions) > 0:
            generate_intrinsic_defintion_files(intrinsic_definitions, args.output, args.use_comments)

    main(sys.argv)


# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import sys
import argparse
from Intrinsic_utils import file_path, dir_path
from Intrinsic_definition_translation import generate_type_definitions_from_modules
from Intrinsic_generator import generate_intrinsic_defintion_files

if __name__ == '__main__':
    def main(args):
        parser = argparse.ArgumentParser(description='Generates IGC intrinsic files.')
        parser.add_argument("inputs", nargs='+', help="the source path to the file with intrinsic defintions (current IGC format)",
                        type=file_path)
        parser.add_argument("--use_comments", action='store_true')
        parser.add_argument("--output", help="the directory for the files with intrinsic definitions",
                        type=dir_path)

        args = parser.parse_args(args[1:])

        intrinsic_definitions = generate_type_definitions_from_modules(args.inputs)
        generate_intrinsic_defintion_files(intrinsic_definitions, args.output, args.use_comments)

    main(sys.argv)


# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

# Wrapper around linker command that filters out LLVM libraries
# for plugin linking.

from sys import argv
from subprocess import check_call

def is_llvm_library(arg):
    if 'LLVMGenXIntrinsics' in arg:
        return False
    if 'LLVM' in arg:
        return True
    return False

args = [arg for arg in argv[1:] if not is_llvm_library(arg)]
check_call(args)



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import argparse
import os
from functools import total_ordering
from subprocess import check_call

parser = argparse.ArgumentParser(description='Apply LLVM patches for IGC')
parser.add_argument('--llvm-version', required=True,
                    help='LLVM version for patching')
parser.add_argument('--llvm-project-dir', required=True,
                    help='LLVM project copied sources')
parser.add_argument('--patches-dir', required=True,
                    help='Directory with patches')
parser.add_argument('--patch-executable', required=True,
                    help='Path to patch program')
parser.add_argument('--patch-disable', required=True,
                    help='Patch to disable')
parser.add_argument('--dry-run', action='store_true',
                    help='Only print list of patches that will be applied')

args = parser.parse_args()

def get_dir_patches(ver_dir):
    """Return list of patches from given version directory.

    Collect patches from 'patches' subdirectory and return them
    as a list of DirEntry objects.
    """
    patches_dir = os.path.join(args.patches_dir, ver_dir)
    patches = []
    ext_patches = os.path.join(patches_dir, 'patches_external')
    if os.path.exists(ext_patches):
        patches.extend(os.scandir(ext_patches))
    return patches

def apply_patch(patch_path):
    """Apply patch to llvm project."""
    rel_path = os.path.relpath(patch_path.path, args.patches_dir)
    print('Applying {} file'.format(rel_path))
    if args.dry_run:
        return
    check_call([args.patch_executable,
                '-t', '-s', '-N',
                '-d', args.llvm_project_dir,
                '-p1', '-i', patch_path])

def split_ver_str(ver_str):
    """Split version string into numeric components.

    Return list of components as numbers additionally checking
    that all components are correct (i.e. can be converted to numbers).
    """
    ver_list = []
    for c in ver_str.split('.')[0:3]:
        if not c.isdecimal():
            raise RuntimeError("Malformed version string '{}'. "
                               "Component '{}' is not an integer."
                               .format(ver_str, c))
        ver_list.append(int(c))

    return ver_list

def get_ver_component(ver_list, idx):
    """Get version component from components list.

    Return 0 for components out of range as default.
    """
    if idx < len(ver_list):
        return ver_list[idx]
    return 0

@total_ordering
class Version:
    """Simple wrapper around three-component version.

    This class provides convenient accessors to version components
    represented by decimal numbers. Suitable for LLVM version in IGC.
    """

    def __init__(self, ver_str):
        ver_list = split_ver_str(ver_str)
        ver_tuple = tuple(get_ver_component(ver_list, idx) for idx in (0, 1, 2))
        self.major = ver_tuple[0]
        self.minor = ver_tuple[1]
        self.patch = ver_tuple[2]

    def as_tuple(self):
        return (self.major, self.minor, self.patch)

    def __repr__(self):
        return '{}.{}.{}'.format(self.major, self.minor, self.patch)

    def __eq__(self, other):
        return self.as_tuple() == other.as_tuple()

    def __lt__(self, other):
        return self.as_tuple() < other.as_tuple()

required_ver = Version(args.llvm_version)

# Create mapping of version to directory with suitable
# version range sorted in descending order.
# All directories with same major version as required are suitable.
patches_dir = os.listdir(args.patches_dir)
ver_to_dir = [(Version(v), v) for v in patches_dir]
ver_to_dir = filter(lambda tpl: tpl[0].major == required_ver.major, ver_to_dir)
ver_to_dir = filter(lambda tpl: tpl[0] <= required_ver, ver_to_dir)
ver_to_dir = sorted(ver_to_dir, key=lambda tpl: tpl[0], reverse=True)

# Merge patches from suitable directories taking only patches from
# newest versions if same patch is present in several directries.
patches = {}
for _, d in ver_to_dir:
    dir_patches = get_dir_patches(d)
    dir_patches = {d.name : d for d in dir_patches}
    dir_patches.update(patches)
    patches = dir_patches
patches = list(patches.values())
patches.sort(key=lambda p: p.name)

checkDisabledPatch = False
if args.patch_disable != "None":
    checkDisabledPatch = True

for patch in patches:
    if checkDisabledPatch:
        if args.patch_disable in patch.name:
            continue
    apply_patch(patch)



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

# -*- Python -*-

import lit.formats
import lit.util

from lit.llvm import llvm_config
from lit.llvm.subst import ToolSubst
from lit.llvm.subst import FindTool

# Configuration file for the 'lit' test runner.

# name: The name of this test suite.
config.name = 'OfflineCompilationTests'

# testFormat: The test format to use to interpret tests.
config.test_format = lit.formats.ShTest(not llvm_config.use_lit_shell)

# suffixes: A list of file extensions to treat as test files.
config.suffixes = ['.cl', '.ll', '.spvasm']

# test_source_root: The root path where tests are located.
config.test_source_root = os.path.dirname(__file__)

# test_exec_root: The root path where tests should be run.
config.test_exec_root = os.path.join(config.test_run_dir, 'test_output')

llvm_config.use_default_substitutions()

llvm_config.with_environment('LD_LIBRARY_PATH', [config.ocloc_lib_dir,
                                                 config.igc_lib_dir,
                                                 config.cclang_lib_dir], append_path=True)


tool_dirs = [config.ocloc_dir, config.llvm_tools_dir, config.spirv_as_dir]

if llvm_config.add_tool_substitutions([ToolSubst('ocloc', unresolved='break')], tool_dirs) is False:
  lit_config.note('Did not find ocloc in %s, ocloc will be used from system paths' % tool_dirs)

llvm_config.add_tool_substitutions([ToolSubst('llvm-dwarfdump')], tool_dirs)
llvm_config.add_tool_substitutions([ToolSubst('opt')], tool_dirs)

if not config.regkeys_disabled:
  config.available_features.add('regkeys')

if config.spirv_as_enabled:
  config.available_features.add('spirv-as')
  llvm_config.add_tool_substitutions([ToolSubst('spirv-as', unresolved='fatal')], tool_dirs)

if config.is32b == "1":
  config.available_features.add('sys32')

if config.use_khronos_spirv_translator_in_sc == "1":
  config.available_features.add('khronos-translator')
  config.available_features.add('khronos-translator-' + config.llvm_version_major)
  config.substitutions.append(('%SPV_CHECK_PREFIX%', 'CHECK-KHR'))
else:
  config.available_features.add('legacy-translator')
  config.substitutions.append(('%SPV_CHECK_PREFIX%', 'CHECK-LEGACY'))



# ========================== begin_copyright_notice ============================
#
# Copyright (C) 2023 Intel Corporation
#
# SPDX-License-Identifier: MIT
#
# =========================== end_copyright_notice =============================

import os
import sys
import re
import json
import importlib.util
import argparse
from Intrinsic_definition_objects import *
from Intrinsic_utils import file_path, Path

def translate_type_definition(type_description):
    if type_description == None:
        return
    if isinstance(type_description, int):
        return TypeDefinition(TypeID.ArgumentReference, index=type_description)
    if isinstance(type_description, list):
        internal_types = []
        for type_str in type_description:
            internal_types.append(translate_type_definition(type_str))
        return TypeDefinition(TypeID.Struct, internal_types=internal_types)
    elif type_description.startswith('any'):
        if type_description == 'anyvector':
            return TypeDefinition(TypeID.Vector, internal_type=TypeDefinition(TypeID.Any))
        elif type_description == 'anyint':
            return TypeDefinition(TypeID.Integer)
        elif type_description == 'anyfloat':
            return TypeDefinition(TypeID.Float)
        elif type_description == 'anyptr':
            return TypeDefinition(TypeID.Pointer, internal_type=TypeDefinition(TypeID.Any))
        else:
            default_type = translate_type_definition(re.search("any(:(.*))?", type_description).group(2))
            return TypeDefinition(TypeID.Any, internal_type=default_type)
    elif type_description == 'void':
        return TypeDefinition(TypeID.Void)
    elif type_description.startswith('bool'):
        type_def = TypeDefinition(TypeID.Integer, bit_width=1)
        match = re.search("bool([0-9]+)", type_description)
        if match is not None:
            type_def = TypeDefinition(TypeID.Vector, num_elements=int(match.group(1)), internal_type=type_def)
        return type_def
    elif type_description.startswith('char'):
        type_def = TypeDefinition(TypeID.Integer, bit_width=8)
        match = re.search("char([0-9]+)", type_description)
        if match is not None:
            type_def = TypeDefinition(TypeID.Vector, num_elements=int(match.group(1)), internal_type=type_def)
        return type_def
    elif type_description.startswith('short'):
        type_def = TypeDefinition(TypeID.Integer, bit_width=16)
        match = re.search("short([0-9]+)", type_description)
        if match is not None:
            type_def = TypeDefinition(TypeID.Vector, num_elements=int(match.group(1)), internal_type=type_def)
        return type_def
    elif type_description.startswith('int'):
        type_def = TypeDefinition(TypeID.Integer, bit_width=32)
        match = re.search("int([0-9]+)", type_description)
        if match is not None:
            type_def = TypeDefinition(TypeID.Vector, num_elements=int(match.group(1)), internal_type=type_def)
        return type_def
    elif type_description.startswith('long'):
        type_def = TypeDefinition(TypeID.Integer, bit_width=64)
        match = re.search("long([0-9]+)", type_description)
        if match is not None:
            type_def = TypeDefinition(TypeID.Vector, num_elements=int(match.group(1)), internal_type=type_def)
        return type_def
    elif type_description.startswith('half'):
        type_def = TypeDefinition(TypeID.Float, bit_width=16)
        match = re.search("half([0-9]+)", type_description)
        if match is not None:
            type_def = TypeDefinition(TypeID.Vector, num_elements=int(match.group(1)), internal_type=type_def)
        return type_def
    elif type_description.startswith('float'):
        type_def = TypeDefinition(TypeID.Float, bit_width=32)
        match = re.search("float([0-9]+)", type_description)
        if match is not None:
            type_def = TypeDefinition(TypeID.Vector, num_elements=int(match.group(1)), internal_type=type_def)
        return type_def
    elif type_description.startswith('double'):
        type_def = TypeDefinition(TypeID.Float, bit_width=32)
        match = re.search("double([0-9]+)", type_description)
        if match is not None:
            type_def = TypeDefinition(TypeID.Vector, num_elements=int(match.group(1)), internal_type=type_def)
        return type_def
    elif type_description.startswith('ptr_private'):
        # E2 <- E == IIT_PTR + AddressSpace == 0 (implicit for IIT_PTR) + 2 == IIT_I8
        return TypeDefinition(TypeID.Pointer, address_space=AddressSpace.Private,
                                       internal_type=TypeDefinition(TypeID.Integer, bit_width=8))
    elif type_description == 'ptr_global':
        # <27>12 (inefficient encoding) <- 27 == IIT_ANYPTR + AddressSpace == 1 (explicit for IIT_ANYPTR) + 2 == IIT_I8
        return TypeDefinition(TypeID.Pointer, address_space=AddressSpace.Global,
                                       internal_type=TypeDefinition(TypeID.Integer, bit_width=8))
    elif type_description == 'ptr_constant':
        # <27>22 (inefficient encoding) <- 27 == IIT_ANYPTR + AddressSpace == 2 (explicit for IIT_ANYPTR) + 2 == IIT_I8
        return TypeDefinition(TypeID.Pointer, address_space=AddressSpace.Constant,
                                       internal_type=TypeDefinition(TypeID.Integer, bit_width=8))
    elif type_description == 'ptr_local':
        # <27>32 (inefficient encoding) <- 27 == IIT_ANYPTR + AddressSpace == 3 (explicit for IIT_ANYPTR) + 2 == IIT_I8
        return TypeDefinition(TypeID.Pointer, address_space=AddressSpace.Local,
                                       internal_type=TypeDefinition(TypeID.Integer, bit_width=8))
    elif type_description == 'ptr_generic':
        # <27>42 (inefficient encoding) <- 27 == IIT_ANYPTR + AddressSpace == 4 (explicit for IIT_ANYPTR) + 2 == IIT_I8
        return TypeDefinition(TypeID.Pointer, address_space=AddressSpace.Generic,
                                       internal_type=TypeDefinition(TypeID.Integer, bit_width=8))
    else:
        return None

def translate_attribute_list(attribute):
    if ',' in attribute:
        attributes = attribute.split(",")
        return set(ID for attribute in attributes for ID in translate_attribute_list(attribute))
    attribute_map = {
        "None": set([ AttributeID.NoUnwind ]),
        "NoMem": set([ AttributeID.NoUnwind, AttributeID.ReadNone ]),
        "ReadMem": set([ AttributeID.NoUnwind, AttributeID.ReadOnly ]),
        "ReadArgMem": set([ AttributeID.NoUnwind, AttributeID.ArgMemOnly, AttributeID.ReadOnly ]),
        "WriteArgMem": set([ AttributeID.NoUnwind, AttributeID.WriteOnly, AttributeID.ArgMemOnly ]),
        "WriteMem": set([ AttributeID.NoUnwind, AttributeID.WriteOnly ]),
        "ReadWriteArgMem": set([ AttributeID.NoUnwind, AttributeID.ArgMemOnly ]),
        "NoReturn": set([ AttributeID.NoUnwind, AttributeID.NoReturn ]),
        "NoDuplicate": set([ AttributeID.NoUnwind, AttributeID.NoDuplicate ]),
        "Convergent": set([ AttributeID.NoUnwind, AttributeID.Convergent ]),
        "InaccessibleMemOnly": set([ AttributeID.NoUnwind, AttributeID.InaccessibleMemOnly ])
    }
    return attribute_map[attribute]

def generate_type_definitions_from_modules(inputs):
    intrinsics = dict()
    for el in inputs:
        spec = importlib.util.spec_from_file_location(Path(el).stem, el)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        intrinsics.update(module.Imported_Intrinsics)

    intrinsic_definitions = []
    for key, value in intrinsics.items():
        name = key
        comment = value[0]
        func_type_def = value[1]
        return_type_str = func_type_def[0]
        if isinstance(return_type_str, list):
            type_strs = [x[0] for x in return_type_str]
            comments = '\n'.join(['Member {}: {}'.format(idx, x[1]) for idx, x in enumerate(return_type_str) if x[1] != ''])
            return_type =  ArgumentTypeDefinition(translate_type_definition(type_strs), comments)
        else:
            return_type =  ArgumentTypeDefinition(translate_type_definition(return_type_str[0]), return_type_str[1])
        argument_type_strs = func_type_def[1]
        argument_types = []
        for type_str in argument_type_strs:
            argument_types.append(ArgumentTypeDefinition(translate_type_definition(type_str[0]), type_str[1]))
        attributes = translate_attribute_list(func_type_def[2])
        intrinsic_definitions.append(IntrinsicDefinition(name, comment, return_type, argument_types, attributes))
    return intrinsic_definitions

if __name__ == '__main__':
    def main(args):
        parser = argparse.ArgumentParser(description='Translate from past IGC format to new IGC format.')
        parser.add_argument("inputs", nargs='+', help="the source path to the file with intrinsic defintions (past IGC format)",
                        type=file_path)
        parser.add_argument("--output", help="the destination path for the file with intrinsic definitions (current IGC format)",
                        type=str)
        parser.add_argument("-v", "--verbose", help="print intrinsic definitions in the current IGC format to the console",
                        action="store_true")

        args = parser.parse_args(args)

        intrinsic_definitions = generate_type_definitions_from_modules(args.inputs)
        serializable_obj = [el.to_dict() for el in intrinsic_definitions]
        ret_json_txt = json.dumps(serializable_obj, indent=2)

        if args.verbose:
            print(ret_json_txt)

        if args.output:
            with open(args.output, 'w') as f:
                f.write(ret_json_txt)

    main(sys.argv)